import pytest

from careamics.config.validators import (
    check_axes_validity,
    patch_size_ge_than_8_power_of_2,
)


@pytest.mark.parametrize(
    "axes, valid",
    [
        # Passing
        ("yx", True),
        ("Yx", True),
        ("Zyx", True),
        ("STYX", True),
        ("CYX", True),
        ("YXC", True),
        ("TzYX", True),
        ("SZYX", True),
        ("STZYX", True),
        ("XY", True),
        ("YXT", True),
        ("ZTYX", True),
        # non consecutive XY
        ("YZX", False),
        ("YZCXT", False),
        # too few axes
        ("", False),
        ("X", False),
        # no yx axes
        ("ZT", False),
        ("ZY", False),
        # repeating characters
        ("YYX", False),
        ("YXY", False),
        # invalid characters
        ("YXm", False),
        ("1YX", False),
    ],
)
def test_are_axes_valid(axes, valid):
    """Test if axes are valid"""
    if valid:
        check_axes_validity(axes)
    else:
        with pytest.raises((ValueError, NotImplementedError)):
            check_axes_validity(axes)


@pytest.mark.parametrize(
    "patch_size, error",
    [
        ((2, 8, 8), True),
        ((10,), True),
        ((8, 10, 16), True),
        ((8, 13), True),
        ((8, 16, 4), True),
        ((8,), False),
        ((8, 8), False),
        ((8, 64, 64), False),
    ],
)
def test_patch_size(patch_size, error):
    """Test if patch size is valid."""
    if error:
        with pytest.raises(ValueError):
            patch_size_ge_than_8_power_of_2(patch_size)
    else:
        patch_size_ge_than_8_power_of_2(patch_size)



"""
Metrics submodule.

This module contains various metrics and a metrics tracking class.
"""

from typing import Callable, Optional, Union

import numpy as np
import torch
from skimage.metrics import peak_signal_noise_ratio, structural_similarity
from torchmetrics.image import MultiScaleStructuralSimilarityIndexMeasure

# TODO: does this add additional dependency?


def psnr(gt: np.ndarray, pred: np.ndarray, data_range: float) -> float:
    """
    Peak Signal to Noise Ratio.

    This method calls skimage.metrics.peak_signal_noise_ratio. See:
    https://scikit-image.org/docs/dev/api/skimage.metrics.html.

    NOTE: to avoid unwanted behaviors (e.g., data_range inferred from array dtype),
    the data_range parameter is mandatory.

    Parameters
    ----------
    gt : np.ndarray
        Ground truth array.
    pred : np.ndarray
        Predicted array.
    data_range : float
        The images pixel range.

    Returns
    -------
    float
        PSNR value.
    """
    return peak_signal_noise_ratio(gt, pred, data_range=data_range)


def _zero_mean(x: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    """
    Zero the mean of an array.

    Parameters
    ----------
    x : numpy.ndarray or torch.Tensor
        Input array.

    Returns
    -------
    numpy.ndarray or torch.Tensor
        Zero-mean array.
    """
    return x - x.mean()


def _fix_range(
    gt: Union[np.ndarray, torch.Tensor], x: Union[np.ndarray, torch.Tensor]
) -> Union[np.ndarray, torch.Tensor]:
    """
    Adjust the range of an array based on a reference ground-truth array.

    Parameters
    ----------
    gt : Union[np.ndarray, torch.Tensor]
        Ground truth array.
    x : Union[np.ndarray, torch.Tensor]
        Input array.

    Returns
    -------
    Union[np.ndarray, torch.Tensor]
        Range-adjusted array.
    """
    a = (gt * x).sum() / (x * x).sum()
    return x * a


def _fix(
    gt: Union[np.ndarray, torch.Tensor], x: Union[np.ndarray, torch.Tensor]
) -> Union[np.ndarray, torch.Tensor]:
    """
    Zero mean a groud truth array and adjust the range of the array.

    Parameters
    ----------
    gt : Union[np.ndarray, torch.Tensor]
        Ground truth image.
    x : Union[np.ndarray, torch.Tensor]
        Input array.

    Returns
    -------
    Union[np.ndarray, torch.Tensor]
        Zero-mean and range-adjusted array.
    """
    gt_ = _zero_mean(gt)
    return _fix_range(gt_, _zero_mean(x))


def scale_invariant_psnr(
    gt: np.ndarray, pred: np.ndarray
) -> Union[float, torch.tensor]:
    """
    Scale invariant PSNR.

    Parameters
    ----------
    gt : np.ndarray
        Ground truth image.
    pred : np.ndarray
        Predicted image.

    Returns
    -------
    Union[float, torch.tensor]
        Scale invariant PSNR value.
    """
    range_parameter = (np.max(gt) - np.min(gt)) / np.std(gt)
    gt_ = _zero_mean(gt) / np.std(gt)
    return psnr(_zero_mean(gt_), _fix(gt_, pred), range_parameter)


class RunningPSNR:
    """Compute the running PSNR during validation step in training.

    This class allows to compute the PSNR on the entire validation set
    one batch at the time.

    Attributes
    ----------
    N : int
        Number of elements seen so far during the epoch.
    mse_sum : float
        Running sum of the MSE over the N elements seen so far.
    max : float
        Running max value of the N target images seen so far.
    min : float
        Running min value of the N target images seen so far.
    """

    def __init__(self):
        """Constructor."""
        self.N = None
        self.mse_sum = None
        self.max = self.min = None
        self.reset()

    def reset(self):
        """Reset the running PSNR computation.

        Usually called at the end of each epoch.
        """
        self.mse_sum = 0
        self.N = 0
        self.max = self.min = None

    def update(self, rec: torch.Tensor, tar: torch.Tensor) -> None:
        """Update the running PSNR statistics given a new batch.

        Parameters
        ----------
        rec : torch.Tensor
            Reconstructed batch.
        tar : torch.Tensor
            Target batch.
        """
        ins_max = torch.max(tar).item()
        ins_min = torch.min(tar).item()
        if self.max is None:
            assert self.min is None
            self.max = ins_max
            self.min = ins_min
        else:
            self.max = max(self.max, ins_max)
            self.min = min(self.min, ins_min)

        mse = (rec - tar) ** 2
        elementwise_mse = torch.mean(mse.view(len(mse), -1), dim=1)
        self.mse_sum += torch.nansum(elementwise_mse)
        self.N += len(elementwise_mse) - torch.sum(torch.isnan(elementwise_mse))

    def get(self) -> Optional[torch.Tensor]:
        """Get the actual PSNR value given the running statistics.

        Returns
        -------
        Optional[torch.Tensor]
            PSNR value.
        """
        if self.N == 0 or self.N is None:
            return None
        rmse = torch.sqrt(self.mse_sum / self.N)
        return 20 * torch.log10((self.max - self.min) / rmse)


def _range_invariant_multiscale_ssim(
    gt_: Union[np.ndarray, torch.Tensor], pred_: Union[np.ndarray, torch.Tensor]
) -> float:
    """Compute range invariant multiscale SSIM for a single channel.

    The advantage of this metric in comparison to commonly used SSIM is that
    it is invariant to scalar multiplications in the prediction.
    # TODO: Add reference to the paper.

    NOTE: images fed to this function should have channels dimension as the last one.

    Parameters
    ----------
    gt_ : Union[np.ndarray, torch.Tensor]
        Ground truth image with shape (N, H, W).
    pred_ : Union[np.ndarray, torch.Tensor]
        Predicted image with shape (N, H, W).

    Returns
    -------
    float
        Range invariant multiscale SSIM value.
    """
    shape = gt_.shape
    gt_ = torch.Tensor(gt_.reshape((shape[0], -1)))
    pred_ = torch.Tensor(pred_.reshape((shape[0], -1)))
    gt_ = _zero_mean(gt_)
    pred_ = _zero_mean(pred_)
    pred_ = _fix(gt_, pred_)
    pred_ = pred_.reshape(shape)
    gt_ = gt_.reshape(shape)

    ms_ssim = MultiScaleStructuralSimilarityIndexMeasure(
        data_range=gt_.max() - gt_.min()
    )
    return ms_ssim(torch.Tensor(pred_[:, None]), torch.Tensor(gt_[:, None])).item()


def multiscale_ssim(
    gt_: Union[np.ndarray, torch.Tensor],
    pred_: Union[np.ndarray, torch.Tensor],
    range_invariant: bool = True,
) -> list[Union[float, None]]:
    """Compute channel-wise multiscale SSIM for each channel.

    It allows to use either standard multiscale SSIM or its range-invariant version.

    NOTE: images fed to this function should have channels dimension as the last one.
    # TODO: do we want to allow this behavior? or we want the usual (N, C, H, W)?

    Parameters
    ----------
    gt_ : Union[np.ndarray, torch.Tensor]
        Ground truth image with shape (N, H, W, C).
    pred_ : Union[np.ndarray, torch.Tensor]
        Predicted image with shape (N, H, W, C).
    range_invariant : bool
        Whether to use standard or range invariant multiscale SSIM.

    Returns
    -------
    list[float]
        List of SSIM values for each channel.
    """
    ms_ssim_values = {}
    for ch_idx in range(gt_.shape[-1]):
        tar_tmp = gt_[..., ch_idx]
        pred_tmp = pred_[..., ch_idx]
        if range_invariant:
            ms_ssim_values[ch_idx] = _range_invariant_multiscale_ssim(
                gt_=tar_tmp, pred_=pred_tmp
            )
        else:
            ms_ssim = MultiScaleStructuralSimilarityIndexMeasure(
                data_range=tar_tmp.max() - tar_tmp.min()
            )
            ms_ssim_values[ch_idx] = ms_ssim(
                torch.Tensor(pred_tmp[:, None]), torch.Tensor(tar_tmp[:, None])
            ).item()

    return [ms_ssim_values[i] for i in range(gt_.shape[-1])]  # type: ignore


def _avg_psnr(target: np.ndarray, prediction: np.ndarray, psnr_fn: Callable) -> float:
    """Compute the average PSNR over a batch of images.

    Parameters
    ----------
    target : np.ndarray
        Array of ground truth images, shape is (N, C, H, W).
    prediction : np.ndarray
        Array of predicted images, shape is (N, C, H, W).
    psnr_fn : Callable
        PSNR function to use.

    Returns
    -------
    float
        Average PSNR value over the batch.
    """
    return np.mean(
        [
            psnr_fn(target[i : i + 1], prediction[i : i + 1]).item()
            for i in range(len(prediction))
        ]
    )


def avg_range_inv_psnr(target: np.ndarray, prediction: np.ndarray) -> float:
    """Compute the average range-invariant PSNR over a batch of images.

    Parameters
    ----------
    target : np.ndarray
        Array of ground truth images, shape is (N, C, H, W).
    prediction : np.ndarray
        Array of predicted images, shape is (N, C, H, W).

    Returns
    -------
    float
        Average range-invariant PSNR value over the batch.
    """
    return _avg_psnr(target, prediction, scale_invariant_psnr)


def avg_psnr(target: np.ndarray, prediction: np.ndarray) -> float:
    """Compute the average PSNR over a batch of images.

    Parameters
    ----------
    target : np.ndarray
        Array of ground truth images, shape is (N, C, H, W).
    prediction : np.ndarray
        Array of predicted images, shape is (N, C, H, W).

    Returns
    -------
    float
        Average PSNR value over the batch.
    """
    return _avg_psnr(target, prediction, psnr)


def avg_ssim(
    target: Union[np.ndarray, torch.Tensor], prediction: Union[np.ndarray, torch.Tensor]
) -> tuple[float, float]:
    """Compute the average Structural Similarity (SSIM) over a batch of images.

    Parameters
    ----------
    target : np.ndarray
        Array of ground truth images, shape is (N, C, H, W).
    prediction : np.ndarray
        Array of predicted images, shape is (N, C, H, W).

    Returns
    -------
    tuple[float, float]
        Mean and standard deviation of SSIM values over the batch.
    """
    ssim = [
        structural_similarity(
            target[i], prediction[i], data_range=(target[i].max() - target[i].min())
        )
        for i in range(len(target))
    ]
    return np.mean(ssim), np.std(ssim)



"""Module containing references to the algorithm used in CAREamics."""

__all__ = [
    "N2V2Ref",
    "N2VRef",
    "StructN2VRef",
    "N2VDescription",
    "N2V2Description",
    "StructN2VDescription",
    "StructN2V2Description",
    "N2V",
    "N2V2",
    "STRUCT_N2V",
    "STRUCT_N2V2",
    "CUSTOM",
    "N2N",
    "CARE",
    "CAREDescription",
    "N2NDescription",
    "CARERef",
    "N2NRef",
]

from .algorithm_descriptions import (
    CARE,
    CUSTOM,
    N2N,
    N2V,
    N2V2,
    STRUCT_N2V,
    STRUCT_N2V2,
    CAREDescription,
    N2NDescription,
    N2V2Description,
    N2VDescription,
    StructN2V2Description,
    StructN2VDescription,
)
from .references import (
    CARERef,
    N2NRef,
    N2V2Ref,
    N2VRef,
    StructN2VRef,
)



import pytest

from careamics.dataset.tiling import collate_tiles, extract_tiles


@pytest.mark.parametrize("n_channels", [1, 4])
@pytest.mark.parametrize("batch", [1, 3])
def test_collate_tiles_2d(ordered_array, n_channels, batch):
    """Test that the collate tiles function collates tile information correctly."""
    tile_size = (4, 4)
    tile_overlap = (2, 2)
    shape = (1, n_channels, 8, 8)

    # create array
    array = ordered_array(shape)

    # extract tiles
    tiles = list(extract_tiles(array, tile_size=tile_size, overlaps=tile_overlap))

    tiles_used = 0
    n_tiles = len(tiles)
    while tiles_used < n_tiles:
        # get a batch of tiles
        batch_tiles = tiles[tiles_used : tiles_used + batch]
        tiles_used += batch

        # collate the tiles
        collated_tiles = collate_tiles(batch_tiles)

        # check the collated tiles
        assert collated_tiles[0].shape == (batch, n_channels) + tile_size

        # check the tile info
        tile_infos = collated_tiles[1]
        assert len(tile_infos) == batch

        for i, t in enumerate(tile_infos):
            for j in range(i + 1, len(tile_infos)):
                assert t != tile_infos[j]


@pytest.mark.parametrize("n_channels", [1, 4])
@pytest.mark.parametrize("batch", [1, 3])
def test_collate_tiles_3d(ordered_array, n_channels, batch):
    """Test that the collate tiles function collates tile information correctly."""
    tile_size = (4, 4, 4)
    tile_overlap = (2, 2, 2)
    shape = (1, n_channels, 8, 8, 8)

    # create array
    array = ordered_array(shape)

    # extract tiles
    tiles = list(extract_tiles(array, tile_size=tile_size, overlaps=tile_overlap))

    tiles_used = 0
    n_tiles = len(tiles)
    while tiles_used < n_tiles:
        # get a batch of tiles
        batch_tiles = tiles[tiles_used : tiles_used + batch]
        tiles_used += batch

        # collate the tiles
        collated_tiles = collate_tiles(batch_tiles)

        # check the collated tiles
        assert collated_tiles[0].shape == (batch, n_channels) + tile_size

        # check the tile info
        tile_infos = collated_tiles[1]
        assert len(tile_infos) == batch

        for i, t in enumerate(tile_infos):
            for j in range(i + 1, len(tile_infos)):
                assert t != tile_infos[j]



"""Optimizers and schedulers supported by CAREamics."""

from careamics.utils import BaseEnum


class SupportedOptimizer(str, BaseEnum):
    """Supported optimizers.

    Attributes
    ----------
    Adam : str
        Adam optimizer.
    SGD : str
        Stochastic Gradient Descent optimizer.
    """

    # ASGD = "ASGD"
    # Adadelta = "Adadelta"
    # Adagrad = "Adagrad"
    ADAM = "Adam"
    # AdamW = "AdamW"
    ADAMAX = "Adamax"
    # LBFGS = "LBFGS"
    # NAdam = "NAdam"
    # RAdam = "RAdam"
    # RMSprop = "RMSprop"
    # Rprop = "Rprop"
    SGD = "SGD"
    # SparseAdam = "SparseAdam"


class SupportedScheduler(str, BaseEnum):
    """Supported schedulers.

    Attributes
    ----------
    ReduceLROnPlateau : str
        Reduce learning rate on plateau.
    StepLR : str
        Step learning rate.
    """

    # ChainedScheduler = "ChainedScheduler"
    # ConstantLR = "ConstantLR"
    # CosineAnnealingLR = "CosineAnnealingLR"
    # CosineAnnealingWarmRestarts = "CosineAnnealingWarmRestarts"
    # CyclicLR = "CyclicLR"
    # ExponentialLR = "ExponentialLR"
    # LambdaLR = "LambdaLR"
    # LinearLR = "LinearLR"
    # MultiStepLR = "MultiStepLR"
    # MultiplicativeLR = "MultiplicativeLR"
    # OneCycleLR = "OneCycleLR"
    # PolynomialLR = "PolynomialLR"
    REDUCE_LR_ON_PLATEAU = "ReduceLROnPlateau"
    # SequentialLR = "SequentialLR"
    STEP_LR = "StepLR"



"""Class representing the parameters of structN2V masks."""

from dataclasses import dataclass
from typing import Literal


@dataclass
class StructMaskParameters:
    """Parameters of structN2V masks.

    Attributes
    ----------
    axis : Literal[0, 1]
        Axis along which to apply the mask, horizontal (0) or vertical (1).
    span : int
        Span of the mask.
    """

    axis: Literal[0, 1]
    span: int



import numpy as np
import pytest

from careamics.dataset.patching.validate_patch_dimension import (
    validate_patch_dimensions,
)


@pytest.mark.parametrize(
    "arr_shape, patch_size",
    [
        ((1, 1, 8, 8), (2, 2)),
        ((1, 1, 8, 8, 8), (2, 2, 2)),
    ],
)
def test_patches_sanity_check(arr_shape, patch_size):
    arr = np.zeros(arr_shape)
    is_3d_patch = len(patch_size) == 3
    # check if the patch is 2D or 3D. Subtract 1 because the first dimension is sample
    validate_patch_dimensions(arr, patch_size, is_3d_patch)


@pytest.mark.parametrize(
    "arr_shape, patch_size",
    [
        # Wrong number of dimensions 2D
        # minimum 3 dimensions CYX
        ((10, 10), (5, 5, 5)),
        # Wrong number of dimensions 3D
        ((1, 1, 10, 10, 10), (5, 5)),
        # Wrong z patch size
        ((1, 10, 10), (5, 5, 5)),
        ((10, 10, 10), (10, 5, 5)),
        # Wrong YX patch sizes
        ((1, 10, 10), (12, 5)),
        ((1, 10, 10), (5, 11)),
    ],
)
def test_patches_sanity_check_invalid_cases(arr_shape, patch_size):
    arr = np.zeros(arr_shape)
    is_3d_patch = len(patch_size) == 3
    # check if the patch is 2D or 3D. Subtract 1 because the first dimension is sample
    with pytest.raises(ValueError):
        validate_patch_dimensions(arr, patch_size, is_3d_patch)



import numpy as np
import pytest

from careamics.dataset.dataset_utils.running_stats import compute_normalization_stats
from careamics.transforms import Denormalize, Normalize
from careamics.transforms.normalize import _reshape_stats


@pytest.mark.parametrize("ndim", [3, 4])
def test_reshape_stats(ndim):
    """Test that reshape stats allows a list of float to be broadcasted to a given
    number of dimensions."""
    shape = (4,) + (8,) * (ndim - 1)
    assert len(shape) == ndim

    # create stats
    stats = [(i + 1) for i in range(shape[0])]

    # create test array
    array = np.ones(shape)

    # reshape and check that you can perform simple operations without error
    reshaped = _reshape_stats(stats, ndim)
    assert reshaped.shape == (shape[0],) + (1,) * (ndim - 1)

    mult = array * reshaped
    for i in range(shape[0]):
        assert (mult[i] == 1 * stats[i]).all()

    add = array + reshaped
    for i in range(shape[0]):
        assert (add[i] == 1 + stats[i]).all()

    sub = array - reshaped
    for i in range(shape[0]):
        assert (sub[i] == 1 - stats[i]).all()

    div = array / reshaped
    for i in range(shape[0]):
        assert (div[i] == 1 / stats[i]).all()


@pytest.mark.parametrize("channels", [1, 2])
def test_normalize_denormalize(channels):
    """Test the Normalize transform."""
    # Create data, adding sample dimension for stats computation
    array = np.arange(100 * channels).reshape((1, channels, 10, 10))

    # Compute mean and std
    means, stds = compute_normalization_stats(image=array)

    # Create the transform
    norm = Normalize(
        image_means=means,
        image_stds=stds,
    )

    # Apply the transform, removing the sample dimension
    normalized, *_ = norm(patch=array[0])
    assert np.abs(normalized.mean()) < 0.02
    assert np.abs(normalized.std() - 1) < 0.2

    # Create the denormalize transform
    denorm = Denormalize(
        image_means=means,
        image_stds=stds,
    )

    # Apply the denormalize transform
    denormalized = denorm(patch=normalized[np.newaxis, ...])  # need to add batch dim
    assert np.isclose(denormalized, array, atol=1e-6).all()


# long name sorry
def test_transform_additional_arrays_not_implemented(ordered_array):
    """Test normalize raises not implemented if additional arrays are used"""
    # create inputs
    shape = (2, 2, 5, 5)
    array = ordered_array(shape)
    additional_arrays = {"arr": ordered_array(shape)}

    # Compute mean and std
    means, stds = compute_normalization_stats(image=array)

    # Create the transform
    norm = Normalize(
        image_means=means,
        image_stds=stds,
    )

    with pytest.raises(NotImplementedError):
        norm(array, **additional_arrays)



from fnmatch import fnmatch
from pathlib import Path

import numpy as np
import pytest
import tifffile

from careamics.config.support import SupportedData


def test_extension_pattern_tiff_fnmatch(tmp_path: Path):
    """Test that the TIFF extension is compatible with fnmatch."""
    path = tmp_path / "test.tif"

    # test as str
    assert fnmatch(str(path), SupportedData.get_extension_pattern(SupportedData.TIFF))

    # test as Path
    assert fnmatch(path, SupportedData.get_extension_pattern(SupportedData.TIFF))


def test_extension_pattern_tiff_rglob(tmp_path: Path):
    """Test that the TIFF extension is compatible with Path.rglob."""
    # create text file
    text_path = tmp_path / "test.txt"
    text_path.write_text("test")

    # create image
    path = tmp_path / "test.tif"
    image = np.ones((10, 10))
    tifffile.imwrite(path, image)

    # search for files
    files = list(
        tmp_path.rglob(SupportedData.get_extension_pattern(SupportedData.TIFF))
    )
    assert len(files) == 1
    assert files[0] == path


def test_extension_pattern_custom_fnmatch(tmp_path: Path):
    """Test that the custom extension is compatible with fnmatch."""
    path = tmp_path / "test.czi"

    # test as str
    assert fnmatch(str(path), SupportedData.get_extension_pattern(SupportedData.CUSTOM))

    # test as Path
    assert fnmatch(path, SupportedData.get_extension_pattern(SupportedData.CUSTOM))


def test_extension_pattern_custom_rglob(tmp_path: Path):
    """Test that the custom extension is compatible with Path.rglob."""
    # create text file
    text_path = tmp_path / "test.txt"
    text_path.write_text("test")

    # create image
    path = tmp_path / "test.npy"
    image = np.ones((10, 10))
    np.save(path, image)

    # search for files
    files = list(
        tmp_path.rglob(SupportedData.get_extension_pattern(SupportedData.CUSTOM))
    )
    assert len(files) == 2
    assert set(files) == {path, text_path}


def test_extension_pattern_array_error():
    """Test that the array extension raises NotImplementedError."""
    with pytest.raises(NotImplementedError):
        SupportedData.get_extension_pattern(SupportedData.ARRAY)


def test_extension_pattern_any_error():
    """Test that any extension raises ValueError."""
    with pytest.raises(ValueError):
        SupportedData.get_extension_pattern("some random")


def test_extension_array_error():
    """Test that the array extension raises NotImplementedError."""
    with pytest.raises(NotImplementedError):
        SupportedData.get_extension(SupportedData.ARRAY)


def test_extension_tiff():
    """Test that the tiff extension is .tiff."""
    assert SupportedData.get_extension(SupportedData.TIFF) == ".tiff"


def test_extension_custom_error():
    """Test that the custom extension returns NotImplementedError."""
    with pytest.raises(NotImplementedError):
        SupportedData.get_extension(SupportedData.CUSTOM)


def test_extension_any_error():
    """Test that any extension raises ValueError."""
    with pytest.raises(ValueError):
        SupportedData.get_extension("some random")



"""Functions relating to writing image files of different formats."""

__all__ = [
    "get_write_func",
    "write_tiff",
    "WriteFunc",
    "SupportedWriteType",
]

from .get_func import (
    SupportedWriteType,
    WriteFunc,
    get_write_func,
)
from .tiff import write_tiff



"""Package to house various prediction utilies."""

__all__ = [
    "stitch_prediction",
    "stitch_prediction_single",
    "convert_outputs",
]

from .prediction_outputs import convert_outputs
from .stitch_prediction import stitch_prediction, stitch_prediction_single



"""Receptive field calculation for computing the tile overlap."""

# TODO better docstring and function names
# Adapted from: https://github.com/frgfm/torch-scan

# import math
# import warnings
# from typing import Tuple, Union

# from torch import Tensor, nn
# from torch.nn import Module
# from torch.nn.modules.batchnorm import _BatchNorm
# from torch.nn.modules.conv import _ConvNd, _ConvTransposeNd
# from torch.nn.modules.pooling import (
#     _AdaptiveAvgPoolNd,
#     _AdaptiveMaxPoolNd,
#     _AvgPoolNd,
#     _MaxPoolNd,
# )


# def module_rf(module: Module, inp: Tensor, out: Tensor) -> Tuple[float, float, float]:
#     """Estimate the spatial receptive field of the module.

#     Parameters
#     ----------
#     module : Module
#         Module to estimate the receptive field.
#     inp : Tensor
#         Input tensor.
#     out : Tensor
#         Output tensor.

#     Returns
#     -------
#     Tuple[float, float, float]
#         Receptive field, effective stride and padding.
#     """
#     if isinstance(
#         module,
#         (
#             nn.Identity,
#             nn.Flatten,
#             nn.ReLU,
#             nn.ELU,
#             nn.LeakyReLU,
#             nn.ReLU6,
#             nn.Tanh,
#             nn.Sigmoid,
#             _BatchNorm,
#             nn.Dropout,
#             nn.Linear,
#         ),
#     ):
#         return 1.0, 1.0, 0.0
#     elif isinstance(module, _ConvTransposeNd):
#         return rf_convtransposend(module, inp, out)
#     elif isinstance(module, (_ConvNd, _MaxPoolNd, _AvgPoolNd)):
#         return rf_aggregnd(module, inp, out)
#     elif isinstance(module, (_AdaptiveMaxPoolNd, _AdaptiveAvgPoolNd)):
#         return rf_adaptive_poolnd(module, inp, out)
#     else:
#         warnings.warn(
#             f"Module type not supported: {module.__class__.__name__}", stacklevel=1
#         )
#         return 1.0, 1.0, 0.0


# def rf_convtransposend(
#     module: _ConvTransposeNd, _: Tensor, __: Tensor
# ) -> Tuple[float, float, float]:
#     k = (
#         module.kernel_size[0]
#         if isinstance(module.kernel_size, tuple)
#         else module.kernel_size
#     )
#     s = module.stride[0] if isinstance(module.stride, tuple) else module.stride
#     return -k, 1.0 / s, 0.0


# def rf_aggregnd(
#     module: Union[_ConvNd, _MaxPoolNd, _AvgPoolNd], _: Tensor, __: Tensor
# ) -> Tuple[float, float, float]:
#     k = (
#         module.kernel_size[0]
#         if isinstance(module.kernel_size, tuple)
#         else module.kernel_size
#     )
#     if hasattr(module, "dilation"):
#         d = (
#             module.dilation[0]
#             if isinstance(module.dilation, tuple)
#             else module.dilation
#         )
#         k = d * (k - 1) + 1
#     s = module.stride[0] if isinstance(module.stride, tuple) else module.stride
#     p = module.padding[0] if isinstance(module.padding, tuple) else module.padding
#     return k, s, p  # type: ignore[return-value]


# def rf_adaptive_poolnd(
#     _: Union[_AdaptiveMaxPoolNd, _AdaptiveAvgPoolNd], inp: Tensor, out: Tensor
# ) -> Tuple[int, int, float]:
#     stride = math.ceil(inp.shape[-1] / out.shape[-1])
#     kernel_size = stride
#     padding = (inp.shape[-1] - kernel_size * stride) / 2

#     return kernel_size, stride, padding



import numpy as np
import pytest

from careamics.transforms import XYFlip


def test_p_wrong_values():
    """Test that an error is raised for wrong probability values."""
    with pytest.raises(ValueError):
        XYFlip(p=-1)

    with pytest.raises(ValueError):
        XYFlip(p=2)


def test_no_flip_error():
    """Test that an error is raised if no axis is flippable."""
    with pytest.raises(ValueError):
        XYFlip(flip_x=False, flip_y=False)


@pytest.mark.parametrize(
    "shape",
    [
        # 2D
        (1, 2, 2),
        (2, 2, 2),
        # 3D
        (1, 2, 2, 2),
        (2, 2, 2, 2),
    ],
)
def test_flip_xy(ordered_array, shape):
    """Test flipping for 2D and 3D arrays."""
    # create array
    array: np.ndarray = ordered_array(shape)

    # create augmentation
    aug = XYFlip(p=1, seed=42)
    r = np.random.default_rng(seed=42)

    # potential flips
    axes = [-2, -1]
    flips = [np.flip(array, axis=axis) for axis in axes]

    # apply augmentation 5 times
    for _ in range(4):
        r.random()  # consume random number
        augmented, *_ = aug(array)

        # draw axis
        axis = r.choice(axes)

        assert np.array_equal(augmented, flips[axis])


@pytest.mark.parametrize(
    "shape",
    [
        # 2D
        (1, 2, 2),
        (2, 2, 2),
        # 3D
        (1, 2, 2, 2),
        (2, 2, 2, 2),
    ],
)
def test_additional_arrays_flip_xy(ordered_array, shape):
    """Test flipping for 2D and 3D arrays."""
    # create array
    array: np.ndarray = ordered_array(shape)
    additional_arrays = {"arr": ordered_array(shape)}

    # create augmentation
    aug = XYFlip(p=1, seed=42)
    r = np.random.default_rng(seed=42)

    # potential flips
    axes = [-2, -1]
    flips = [np.flip(array, axis=axis) for axis in axes]

    # apply augmentation 5 times
    for _ in range(4):
        r.random()  # consume random number
        augmented, _, additional_augmented = aug(array, **additional_arrays)

        # draw axis
        axis = r.choice(axes)

        assert np.array_equal(augmented, flips[axis])
        assert np.array_equal(augmented, additional_augmented["arr"])


@pytest.mark.parametrize(
    "shape, flip_x",
    [
        # 2D
        ((1, 2, 2), True),
        ((2, 2, 2), True),
        ((1, 2, 2), False),
        ((2, 2, 2), False),
        # 3D
        ((1, 2, 2), True),
        ((2, 2, 2), True),
        ((1, 2, 2), False),
        ((2, 2, 2), False),
    ],
)
def test_flip_single_axis(ordered_array, shape, flip_x):
    """Test flipping for 2D and 3D arrays."""
    # create array
    array: np.ndarray = ordered_array(shape)

    # create augmentation
    aug = XYFlip(flip_x=flip_x, flip_y=not flip_x, p=1)

    # potential flips
    axis = -1 if flip_x else -2

    # apply augmentation 5 times
    for _ in range(5):
        augmented, *_ = aug(array)

        assert np.array_equal(augmented, np.flip(array, axis=axis))


def test_flip_mask(ordered_array):
    """Test flipping masks in 3D."""
    # create array
    array = ordered_array((4, 2, 2, 2))
    mask = array[2:, ...]
    array = array[:2, ...]

    # create augmentation
    aug = XYFlip(p=1, seed=42)
    r = np.random.default_rng(seed=42)

    # potential flips on Y and X axes
    axes = [-2, -1]
    array_flips = [np.flip(array, axis=axis) for axis in axes]
    mask_flips = [np.flip(mask, axis=axis) for axis in axes]

    # apply augmentation 5 times
    for _ in range(5):
        aug_array, aug_mask, _ = aug(patch=array, target=mask)
        r.random()  # consume random number
        axis = r.choice(axes)

        assert np.array_equal(aug_array, array_flips[axis])
        assert np.array_equal(aug_mask, mask_flips[axis])


def test_p(ordered_array):
    """Test that the probability of flipping is respected."""
    # create array
    array = ordered_array((2, 2, 2))

    # create augmentation that never applies
    aug = XYFlip(p=0.0, seed=42)

    # apply augmentation 5 times
    for _ in range(5):
        augmented, *_ = aug(array)

        assert np.array_equal(augmented, array)

    # create augmentation that always applies
    aug = XYFlip(p=1.0, seed=42)

    # apply augmentation 5 times
    for _ in range(5):
        augmented, *_ = aug(array)

        assert not np.array_equal(augmented, array)



"""Collate function for tiling."""

from typing import Any, List, Tuple

import numpy as np
from torch.utils.data.dataloader import default_collate

from careamics.config.tile_information import TileInformation


def collate_tiles(batch: List[Tuple[np.ndarray, TileInformation]]) -> Any:
    """
    Collate tiles received from CAREamics prediction dataloader.

    CAREamics prediction dataloader returns tuples of arrays and TileInformation. In
    case of non-tiled data, this function will return the arrays. In case of tiled data,
    it will return the arrays, the last tile flag, the overlap crop coordinates and the
    stitch coordinates.

    Parameters
    ----------
    batch : List[Tuple[np.ndarray, TileInformation], ...]
        Batch of tiles.

    Returns
    -------
    Any
        Collated batch.
    """
    new_batch = [tile for tile, _ in batch]
    tiles_batch = [tile_info for _, tile_info in batch]

    return default_collate(new_batch), tiles_batch



"""Callbacks module."""

__all__ = [
    "HyperParametersCallback",
    "ProgressBarCallback",
    "PredictionWriterCallback",
]

from .hyperparameters_callback import HyperParametersCallback
from .prediction_writer_callback import PredictionWriterCallback
from .progress_bar_callback import ProgressBarCallback



from torch import optim

from careamics.config.support.supported_optimizers import (
    SupportedOptimizer,
    SupportedScheduler,
)


def test_schedulers_exist():
    """Test that `SupportedScheduler` contains existing torch schedulers."""
    for scheduler in SupportedScheduler:
        assert hasattr(optim.lr_scheduler, scheduler)


def test_optimizers_exist():
    """Test that `SupportedOptimizer` contains existing torch optimizers."""
    for optimizer in SupportedOptimizer:
        assert hasattr(optim, optimizer)



import pytest

from careamics.config.fcn_algorithm_model import FCNAlgorithmConfig
from careamics.config.support import (
    SupportedAlgorithm,
    SupportedArchitecture,
    SupportedLoss,
)


def test_all_algorithms_are_supported():
    """Test that all algorithms defined in the Literal are supported."""
    # list of supported algorithms
    algorithms = list(SupportedAlgorithm)

    # Algorithm json schema to extract the literal value
    schema = FCNAlgorithmConfig.model_json_schema()

    # check that all algorithms are supported
    for algo in schema["properties"]["algorithm"]["enum"]:
        assert algo in algorithms


# TODO: this should not support musplit and denoisplit losses
def test_all_losses_are_supported():
    """Test that all losses defined in the Literal are supported."""
    # list of supported losses
    losses = list(SupportedLoss)

    # Algorithm json schema
    schema = FCNAlgorithmConfig.model_json_schema()

    # check that all losses are supported
    for loss in schema["properties"]["loss"]["enum"]:
        assert loss in losses


def test_model_discriminator(minimum_algorithm_n2v):
    """Test that discriminator permits correct assignment."""
    for model_name in SupportedArchitecture:
        # TODO change once VAE are implemented
        if model_name.value == "UNet":
            minimum_algorithm_n2v["model"]["architecture"] = model_name.value

            algo = FCNAlgorithmConfig(**minimum_algorithm_n2v)
            assert algo.model.architecture == model_name.value


@pytest.mark.parametrize(
    "algorithm, loss, model",
    [
        ("n2v", "n2v", {"architecture": "UNet", "n2v2": False}),
        ("n2n", "mae", {"architecture": "UNet", "n2v2": False}),
        ("care", "mae", {"architecture": "UNet", "n2v2": False}),
    ],
)
def test_algorithm_constraints(algorithm: str, loss: str, model: dict):
    """Test that constraints are passed for each algorithm."""
    FCNAlgorithmConfig(algorithm=algorithm, loss=loss, model=model)


def test_n_channels_n2v():
    """Check that an error is raised if n2v has different number of channels in
    input and output."""
    model = {
        "architecture": "UNet",
        "in_channels": 1,
        "num_classes": 2,
        "n2v2": False,
    }
    loss = "n2v"

    with pytest.raises(ValueError):
        FCNAlgorithmConfig(algorithm="n2v", loss=loss, model=model)


@pytest.mark.parametrize(
    "algorithm, n_in, n_out",
    [
        ("n2v", 2, 2),
        ("n2n", 3, 3),
        ("care", 1, 2),
    ],
)
def test_comaptiblity_of_number_of_channels(algorithm, n_in, n_out):
    """Check that no error is thrown when instantiating the algorithm with a valid
    number of in and out channels."""
    model = {
        "architecture": "UNet",
        "in_channels": n_in,
        "num_classes": n_out,
        "n2v2": False,
    }
    loss = "n2v" if algorithm == "n2v" else "mae"

    FCNAlgorithmConfig(algorithm=algorithm, loss=loss, model=model)


def test_custom_model(custom_model_parameters):
    """Test that a custom model can be instantiated."""
    # create algorithm configuration
    FCNAlgorithmConfig(
        algorithm=SupportedAlgorithm.CUSTOM.value,
        loss="mse",
        model=custom_model_parameters,
    )


def test_custom_model_wrong_algorithm(custom_model_parameters):
    """Test that a custom model fails if the algorithm is not custom."""
    # create algorithm configuration
    with pytest.raises(ValueError):
        FCNAlgorithmConfig(
            algorithm=SupportedAlgorithm.CARE.value,
            loss="mse",
            model=custom_model_parameters,
        )



"""Utils module."""

__all__ = [
    "cwd",
    "get_ram_size",
    "check_path_exists",
    "BaseEnum",
    "get_logger",
    "get_careamics_home",
    "autocorrelation",
]


from .autocorrelation import autocorrelation
from .base_enum import BaseEnum
from .context import cwd, get_careamics_home
from .logging import get_logger
from .path_utils import check_path_exists
from .ram import get_ram_size



from __future__ import annotations

from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Literal, Union

import numpy as np
import pytest
import torch

from careamics.config import (
    GaussianMixtureNMConfig,
    LVAELossConfig,
    MultiChannelNMConfig,
)
from careamics.config.likelihood_model import (
    GaussianLikelihoodConfig,
    NMLikelihoodConfig,
)
from careamics.config.loss_model import KLLossConfig
from careamics.losses.loss_factory import (
    SupportedLoss,
    loss_factory,
)
from careamics.losses.lvae.losses import (
    _reconstruction_loss_musplit_denoisplit,
    denoisplit_loss,
    denoisplit_musplit_loss,
    get_kl_divergence_loss,
    get_reconstruction_loss,
    musplit_loss,
)
from careamics.models.lvae.likelihoods import likelihood_factory
from careamics.models.lvae.noise_models import noise_model_factory

if TYPE_CHECKING:
    from careamics.models.lvae.noise_models import MultiChannelNoiseModel

pytestmark = pytest.mark.lvae


# TODO: move to conftest.py as pytest.fixture
def create_dummy_noise_model(
    tmp_path: Path,
    n_gaussians: int = 3,
    n_coeffs: int = 3,
) -> None:
    """Create a dummy noise model and save it in a `.npz` file."""
    weights = np.random.rand(3 * n_gaussians, n_coeffs)
    nm_dict = {
        "trained_weight": weights,
        "min_signal": np.array([0]),
        "max_signal": np.array([2**16 - 1]),
        "min_sigma": 0.125,
    }
    np.savez(tmp_path / "dummy_noise_model.npz", **nm_dict)


def init_noise_model(
    tmp_path: Path,
    target_ch: int,
    n_gaussians: int = 3,
    n_coeffs: int = 3,
) -> MultiChannelNoiseModel:
    """Instantiate a dummy noise model."""
    create_dummy_noise_model(tmp_path, n_gaussians, n_coeffs)
    gmm = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        path=tmp_path / "dummy_noise_model.npz",
        # all other params are default
    )
    noise_model_config = MultiChannelNMConfig(noise_models=[gmm] * target_ch)
    return noise_model_factory(noise_model_config)


@pytest.mark.parametrize(
    "loss_type, exp_loss_func, exp_error",
    [
        (SupportedLoss.MUSPLIT, musplit_loss, does_not_raise()),
        (SupportedLoss.DENOISPLIT, denoisplit_loss, does_not_raise()),
        (SupportedLoss.DENOISPLIT_MUSPLIT, denoisplit_musplit_loss, does_not_raise()),
        ("musplit", musplit_loss, does_not_raise()),
        ("denoisplit", denoisplit_loss, does_not_raise()),
        ("denoisplit_musplit", denoisplit_musplit_loss, does_not_raise()),
        ("made_up_loss", None, pytest.raises(NotImplementedError)),
    ],
)
def test_lvae_loss_factory(
    loss_type: Union[SupportedLoss, str], exp_loss_func: Callable, exp_error: Callable
):
    with exp_error:
        loss_func = loss_factory(loss_type)
        assert loss_func is not None
        assert callable(loss_func)
        assert loss_func == exp_loss_func


@pytest.mark.parametrize(
    "batch_size, target_ch, predict_logvar, likelihood_type",
    [
        (1, 1, None, "noise_model"),
        (1, 1, None, "gaussian"),
        (1, 1, "pixelwise", "gaussian"),
        (8, 1, None, "noise_model"),
        (8, 1, None, "gaussian"),
        (8, 1, "pixelwise", "gaussian"),
        (1, 3, None, "noise_model"),
        (1, 3, None, "gaussian"),
        (1, 3, "pixelwise", "gaussian"),
        (8, 3, None, "noise_model"),
        (8, 3, None, "gaussian"),
        (8, 3, "pixelwise", "gaussian"),
    ],
)
def test_reconstruction_loss(
    tmp_path: Path,
    batch_size: int,
    target_ch: int,
    predict_logvar: str,
    likelihood_type: str,
):
    # create test data
    img_size = 64
    inp_ch = target_ch * (1 + int(predict_logvar is not None))
    reconstruction = torch.rand((batch_size, inp_ch, img_size, img_size))
    target = torch.rand((batch_size, target_ch, img_size, img_size))

    # create likelihood object
    if likelihood_type == "noise_model":
        nm = init_noise_model(tmp_path, target_ch)
        data_mean = target.mean(dim=(0, 2, 3), keepdim=True)
        data_std = target.std(dim=(0, 2, 3), keepdim=True)
        config = NMLikelihoodConfig(data_mean=data_mean, data_std=data_std)
    else:
        nm = None
        config = GaussianLikelihoodConfig(predict_logvar=predict_logvar)
    likelihood = likelihood_factory(config, noise_model=nm)

    # compute the loss
    rec_loss = get_reconstruction_loss(
        reconstruction=reconstruction, target=target, likelihood_obj=likelihood
    )

    # check outputs
    assert rec_loss is not None


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("target_ch", [1, 3])
@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("nm_weight", [0.0, 0.5, 1.0])
def test_reconstruction_loss_musplit_denoisplit(
    tmp_path: Path,
    batch_size: int,
    target_ch: int,
    predict_logvar: str,
    nm_weight: float,
):
    # create test data
    img_size = 64
    inp_ch = target_ch * (1 + int(predict_logvar is not None))
    reconstruction = torch.rand((batch_size, inp_ch, img_size, img_size))
    target = torch.rand((batch_size, target_ch, img_size, img_size))

    # create likelihood objects
    nm = init_noise_model(tmp_path, target_ch)
    data_mean = target.mean(dim=(0, 2, 3), keepdim=True)
    data_std = target.std(dim=(0, 2, 3), keepdim=True)
    nm_config = NMLikelihoodConfig(data_mean=data_mean, data_std=data_std)
    nm_likelihood = likelihood_factory(nm_config, noise_model=nm)
    gaussian_config = GaussianLikelihoodConfig(predict_logvar=predict_logvar)
    gaussian_likelihood = likelihood_factory(gaussian_config)

    # compute the loss
    rec_loss = _reconstruction_loss_musplit_denoisplit(
        predictions=reconstruction,
        targets=target,
        nm_likelihood=nm_likelihood,
        gaussian_likelihood=gaussian_likelihood,
        nm_weight=nm_weight,
        gaussian_weight=1.0 - nm_weight,
    )

    # check outputs
    assert rec_loss is not None
    assert isinstance(rec_loss, torch.Tensor)


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("n_layers", [1, 4])
@pytest.mark.parametrize("enable_LC", [False, True])
@pytest.mark.parametrize("kl_type", ["kl", "kl_restricted"])
@pytest.mark.parametrize("rescaling", ["latent_dim", "image_dim"])
@pytest.mark.parametrize("aggregation", ["mean", "sum"])
@pytest.mark.parametrize("free_bits_coeff", [0.0, 1.0])
def test_KL_divergence_loss(
    batch_size: int,
    n_layers: int,
    enable_LC: bool,
    kl_type: Literal["kl", "kl_restricted"],
    rescaling: Literal["latent_dim", "image_dim"],
    aggregation: Literal["mean", "sum"],
    free_bits_coeff: float,
):
    # create test data
    img_size = 64
    if enable_LC:
        z = [torch.ones(batch_size, 128, img_size, img_size) for _ in range(n_layers)]
    else:
        sizes = [img_size // 2 ** (i + 1) for i in range(n_layers)]
        sizes = sizes[::-1]
        z = [torch.ones(batch_size, 128, sz, sz) for sz in sizes]
    td_data = {
        "z": z,
        "kl": [torch.ones(batch_size) for _ in range(n_layers)],
        "kl_restricted": [torch.ones(batch_size) for _ in range(n_layers)],
    }

    # compute the loss for different settings
    img_shape = (img_size, img_size)
    kl_loss = get_kl_divergence_loss(
        kl_type=kl_type,
        topdown_data=td_data,
        rescaling=rescaling,
        aggregation=aggregation,
        free_bits_coeff=free_bits_coeff,
        img_shape=img_shape,
    )
    assert isinstance(kl_loss, torch.Tensor)
    assert isinstance(kl_loss.item(), float)


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("target_ch", [1, 3])
@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("n_layers", [1, 4])
@pytest.mark.parametrize("enable_LC", [False, True])
@pytest.mark.parametrize("kl_type", ["kl", "kl_restricted"])
def test_musplit_loss(
    batch_size: int,
    target_ch: int,
    predict_logvar: str,
    n_layers: int,
    enable_LC: bool,
    kl_type: Literal["kl", "kl_restricted"],
):
    # create test data
    img_size = 64
    inp_ch = target_ch * (1 + int(predict_logvar is not None))
    reconstruction = torch.rand((batch_size, inp_ch, img_size, img_size))
    target = torch.rand((batch_size, target_ch, img_size, img_size))
    if enable_LC:
        z = [torch.rand(batch_size, 128, img_size, img_size) for _ in range(n_layers)]
    else:
        sizes = [img_size // 2 ** (i + 1) for i in range(n_layers)]
        sizes = sizes[::-1]
        z = [torch.rand(batch_size, 128, sz, sz) for sz in sizes]
    td_data = {
        "z": z,
        "kl": [torch.rand(batch_size) for _ in range(n_layers)],
        "kl_restricted": [torch.rand(batch_size) for _ in range(n_layers)],
    }

    # create likelihood
    config = GaussianLikelihoodConfig(predict_logvar=predict_logvar)
    likelihood = likelihood_factory(config)

    # compute the loss
    kl_params = KLLossConfig(loss_type=kl_type)
    loss_parameters = LVAELossConfig(loss_type="musplit", kl_params=kl_params)
    output = musplit_loss(
        model_outputs=(reconstruction, td_data),
        targets=target,
        config=loss_parameters,
        gaussian_likelihood=likelihood,
    )

    # check outputs
    # NOTE: output should not be None in these test cases
    assert output is not None
    assert isinstance(output, dict)
    assert "loss" in output
    assert "reconstruction_loss" in output
    assert "kl_loss" in output


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("target_ch", [1, 3])
@pytest.mark.parametrize("n_layers", [1, 4])
@pytest.mark.parametrize("kl_type", ["kl", "kl_restricted"])
def test_denoisplit_loss(
    tmp_path: Path,
    batch_size: int,
    target_ch: int,
    n_layers: int,
    kl_type: Literal["kl", "kl_restricted"],
):
    # create test data
    img_size = 64
    reconstruction = torch.rand((batch_size, target_ch, img_size, img_size))
    target = torch.rand((batch_size, target_ch, img_size, img_size))
    td_data = {
        "z": [torch.rand(batch_size, 128, img_size, img_size) for _ in range(n_layers)],
        "kl": [torch.rand(batch_size) for _ in range(n_layers)],
        "kl_restricted": [torch.rand(batch_size) for _ in range(n_layers)],
    }

    # create likelihood
    nm = init_noise_model(tmp_path, target_ch)
    data_mean = target.mean(dim=(0, 2, 3), keepdim=True)
    data_std = target.std(dim=(0, 2, 3), keepdim=True)
    nm_config = NMLikelihoodConfig(data_mean=data_mean, data_std=data_std)
    likelihood = likelihood_factory(nm_config, noise_model=nm)

    # compute the loss
    kl_params = KLLossConfig(loss_type=kl_type)
    loss_parameters = LVAELossConfig(loss_type="denoisplit", kl_params=kl_params)
    output = denoisplit_loss(
        model_outputs=(reconstruction, td_data),
        targets=target,
        config=loss_parameters,
        noise_model_likelihood=likelihood,
    )

    # check outputs
    # NOTE: output should not be None in these test cases
    assert output is not None
    assert isinstance(output, dict)
    assert "loss" in output
    assert "reconstruction_loss" in output
    assert "kl_loss" in output


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("target_ch", [1, 3])
@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("n_layers", [1, 4])
@pytest.mark.parametrize("enable_LC", [False, True])
@pytest.mark.parametrize("kl_type", ["kl", "kl_restricted"])
def test_denoisplit_musplit_loss(
    tmp_path: Path,
    batch_size: int,
    target_ch: int,
    predict_logvar: str,
    n_layers: int,
    enable_LC: bool,
    kl_type: Literal["kl", "kl_restricted"],
):
    # create test data
    img_size = 64
    inp_ch = target_ch * (1 + int(predict_logvar is not None))
    reconstruction = torch.rand((batch_size, inp_ch, img_size, img_size))
    target = torch.rand((batch_size, target_ch, img_size, img_size))
    if enable_LC:
        z = [torch.rand(batch_size, 128, img_size, img_size) for _ in range(n_layers)]
    else:
        sizes = [img_size // 2 ** (i + 1) for i in range(n_layers)]
        sizes = sizes[::-1]
        z = [torch.rand(batch_size, 128, sz, sz) for sz in sizes]
    td_data = {
        "z": z,
        "kl": [torch.rand(batch_size) for _ in range(n_layers)],
        "kl_restricted": [torch.rand(batch_size) for _ in range(n_layers)],
    }

    # create likelihood objects
    nm = init_noise_model(tmp_path, target_ch)
    data_mean = target.mean(dim=(0, 2, 3), keepdim=True)
    data_std = target.std(dim=(0, 2, 3), keepdim=True)
    nm_config = NMLikelihoodConfig(data_mean=data_mean, data_std=data_std)
    nm_likelihood = likelihood_factory(nm_config, noise_model=nm)
    gaussian_config = GaussianLikelihoodConfig(predict_logvar=predict_logvar)
    gaussian_likelihood = likelihood_factory(gaussian_config)

    # compute the loss
    kl_params = KLLossConfig(loss_type=kl_type)
    loss_parameters = LVAELossConfig(
        loss_type="denoisplit_musplit", kl_params=kl_params
    )
    output = denoisplit_musplit_loss(
        model_outputs=(reconstruction, td_data),
        targets=target,
        config=loss_parameters,
        gaussian_likelihood=gaussian_likelihood,
        noise_model_likelihood=nm_likelihood,
    )

    # check outputs
    # NOTE: output should not be None in these test cases
    assert output is not None
    assert isinstance(output, dict)
    assert "loss" in output
    assert "reconstruction_loss" in output
    assert "kl_loss" in output



"""Type used to represent all transformations users can create."""

from typing import Union

from pydantic import Discriminator
from typing_extensions import Annotated

from .n2v_manipulate_model import N2VManipulateModel
from .xy_flip_model import XYFlipModel
from .xy_random_rotate90_model import XYRandomRotate90Model

TRANSFORMS_UNION = Annotated[
    Union[
        XYFlipModel,
        XYRandomRotate90Model,
        N2VManipulateModel,
    ],
    Discriminator("name"),  # used to tell the different transform models apart
]
"""Available transforms in CAREamics."""



from torch import nn, ones

from careamics.config.architectures import (
    CustomModel,
    UNetModel,
    register_model,
)
from careamics.config.support import SupportedArchitecture
from careamics.models import model_factory
from careamics.models.unet import UNet


def test_model_registry_unet():
    """Test that"""
    model_config = {
        "architecture": "UNet",
    }

    # instantiate model
    model = model_factory(UNetModel(**model_config))
    assert isinstance(model, UNet)


def test_model_registry_custom():
    """Test that a custom model can be retrieved and instantiated."""

    # create and register a custom model
    @register_model(name="linear_model")
    class LinearModel(nn.Module):
        def __init__(self, in_features, out_features):
            super().__init__()

            self.in_features = in_features
            self.out_features = out_features
            self.weight = nn.Parameter(ones(in_features, out_features))
            self.bias = nn.Parameter(ones(out_features))

        def forward(self, input):
            return (input @ self.weight) + self.bias

    model_config = {
        "architecture": SupportedArchitecture.CUSTOM.value,
        "name": "linear_model",
        "in_features": 10,
        "out_features": 5,
    }

    # instantiate model
    model = model_factory(CustomModel(**model_config))
    assert isinstance(model, LinearModel)
    assert model.in_features == 10
    assert model.out_features == 5



"""Parent model for the transforms."""

from typing import Any, Dict

from pydantic import BaseModel, ConfigDict


class TransformModel(BaseModel):
    """
    Pydantic model used to represent a transformation.

    The `model_dump` method is overwritten to exclude the name field.

    Attributes
    ----------
    name : str
        Name of the transformation.
    """

    model_config = ConfigDict(
        extra="forbid",  # throw errors if the parameters are not properly passed
    )

    name: str

    def model_dump(self, **kwargs) -> Dict[str, Any]:
        """
        Return the model as a dictionary.

        Parameters
        ----------
        **kwargs
            Pydantic BaseMode model_dump method keyword arguments.

        Returns
        -------
        Dict[str, Any]
            Dictionary representation of the model.
        """
        model_dict = super().model_dump(**kwargs)

        # remove the name field
        model_dict.pop("name")

        return model_dict



"""
Loss submodule.

This submodule contains the various losses used in CAREamics.
"""

import torch
from torch.nn import L1Loss, MSELoss


def mse_loss(source: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """
    Mean squared error loss.

    Parameters
    ----------
    source : torch.Tensor
        Source patches.
    target : torch.Tensor
        Target patches.

    Returns
    -------
    torch.Tensor
        Loss value.
    """
    loss = MSELoss()
    return loss(source, target)


def n2v_loss(
    manipulated_patches: torch.Tensor,
    original_patches: torch.Tensor,
    masks: torch.Tensor,
) -> torch.Tensor:
    """
    N2V Loss function described in A Krull et al 2018.

    Parameters
    ----------
    manipulated_patches : torch.Tensor
        Patches with manipulated pixels.
    original_patches : torch.Tensor
        Noisy patches.
    masks : torch.Tensor
        Array containing masked pixel locations.

    Returns
    -------
    torch.Tensor
        Loss value.
    """
    errors = (original_patches - manipulated_patches) ** 2
    # Average over pixels and batch
    loss = torch.sum(errors * masks) / torch.sum(masks)
    return loss  # TODO change output to dict ?


def mae_loss(samples: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
    """
    N2N Loss function described in to J Lehtinen et al 2018.

    Parameters
    ----------
    samples : torch.Tensor
        Raw patches.
    labels : torch.Tensor
        Different subset of noisy patches.

    Returns
    -------
    torch.Tensor
        Loss value.
    """
    loss = L1Loss()
    return loss(samples, labels)


# def pn2v_loss(
#     samples: torch.Tensor,
#     labels: torch.Tensor,
#     masks: torch.Tensor,
#     noise_model: HistogramNoiseModel,
# ) -> torch.Tensor:
#     """Probabilistic N2V loss function described in A Krull et al., CVF (2019)."""
#     likelihoods = noise_model.likelihood(labels, samples)
#     likelihoods_avg = torch.log(torch.mean(likelihoods, dim=0, keepdim=True)[0, ...])

#     # Average over pixels and batch
#     loss = -torch.sum(likelihoods_avg * masks) / torch.sum(masks)
#     return loss


# def dice_loss(
#     samples: torch.Tensor, labels: torch.Tensor, mode: str = "multiclass"
# ) -> torch.Tensor:
#     """Dice loss function."""
#     return DiceLoss(mode=mode)(samples, labels.long())



"""Module to get read functions."""

from pathlib import Path
from typing import Callable, Dict, Protocol, Union

from numpy.typing import NDArray

from careamics.config.support import SupportedData

from .tiff import read_tiff


# This is very strict, function signature has to match including arg names
# See WriteFunc notes
class ReadFunc(Protocol):
    """Protocol for type hinting read functions."""

    def __call__(self, file_path: Path, *args, **kwargs) -> NDArray:
        """
        Type hinted callables must match this function signature (not including self).

        Parameters
        ----------
        file_path : pathlib.Path
            Path to file.
        *args
            Other positional arguments.
        **kwargs
            Other keyword arguments.
        """


READ_FUNCS: Dict[SupportedData, ReadFunc] = {
    SupportedData.TIFF: read_tiff,
}


def get_read_func(data_type: Union[str, SupportedData]) -> Callable:
    """
    Get the read function for the data type.

    Parameters
    ----------
    data_type : SupportedData
        Data type.

    Returns
    -------
    callable
        Read function.
    """
    if data_type in READ_FUNCS:
        data_type = SupportedData(data_type)  # mypy complaining about dict key type
        return READ_FUNCS[data_type]
    else:
        raise NotImplementedError(f"Data type '{data_type}' is not supported.")



"""N2V manipulation transform."""

from typing import Any, Literal, Optional, Tuple

import numpy as np
from numpy.typing import NDArray

from careamics.config.support import SupportedPixelManipulation, SupportedStructAxis
from careamics.transforms.transform import Transform

from .pixel_manipulation import median_manipulate, uniform_manipulate
from .struct_mask_parameters import StructMaskParameters


class N2VManipulate(Transform):
    """
    Default augmentation for the N2V model.

    This transform expects C(Z)YX dimensions.

    Parameters
    ----------
    roi_size : int, optional
        Size of the replacement area, by default 11.
    masked_pixel_percentage : float, optional
        Percentage of pixels to mask, by default 0.2.
    strategy : Literal[ "uniform", "median" ], optional
        Replaccement strategy, uniform or median, by default uniform.
    remove_center : bool, optional
        Whether to remove central pixel from patch, by default True.
    struct_mask_axis : Literal["horizontal", "vertical", "none"], optional
        StructN2V mask axis, by default "none".
    struct_mask_span : int, optional
        StructN2V mask span, by default 5.
    seed : Optional[int], optional
        Random seed, by default None.

    Attributes
    ----------
    masked_pixel_percentage : float
        Percentage of pixels to mask.
    roi_size : int
        Size of the replacement area.
    strategy : Literal[ "uniform", "median" ]
        Replaccement strategy, uniform or median.
    remove_center : bool
        Whether to remove central pixel from patch.
    struct_mask : Optional[StructMaskParameters]
        StructN2V mask parameters.
    rng : Generator
        Random number generator.
    """

    def __init__(
        self,
        roi_size: int = 11,
        masked_pixel_percentage: float = 0.2,
        strategy: Literal[
            "uniform", "median"
        ] = SupportedPixelManipulation.UNIFORM.value,
        remove_center: bool = True,
        struct_mask_axis: Literal["horizontal", "vertical", "none"] = "none",
        struct_mask_span: int = 5,
        seed: Optional[int] = None,
    ):
        """Constructor.

        Parameters
        ----------
        roi_size : int, optional
            Size of the replacement area, by default 11.
        masked_pixel_percentage : float, optional
            Percentage of pixels to mask, by default 0.2.
        strategy : Literal[ "uniform", "median" ], optional
            Replaccement strategy, uniform or median, by default uniform.
        remove_center : bool, optional
            Whether to remove central pixel from patch, by default True.
        struct_mask_axis : Literal["horizontal", "vertical", "none"], optional
            StructN2V mask axis, by default "none".
        struct_mask_span : int, optional
            StructN2V mask span, by default 5.
        seed : Optional[int], optional
            Random seed, by default None.
        """
        self.masked_pixel_percentage = masked_pixel_percentage
        self.roi_size = roi_size
        self.strategy = strategy
        self.remove_center = remove_center  # TODO is this ever used?

        if struct_mask_axis == SupportedStructAxis.NONE:
            self.struct_mask: Optional[StructMaskParameters] = None
        else:
            self.struct_mask = StructMaskParameters(
                axis=0 if struct_mask_axis == SupportedStructAxis.HORIZONTAL else 1,
                span=struct_mask_span,
            )

        # numpy random generator
        self.rng = np.random.default_rng(seed=seed)

    def __call__(
        self, patch: NDArray, *args: Any, **kwargs: Any
    ) -> Tuple[NDArray, NDArray, NDArray]:
        """Apply the transform to the image.

        Parameters
        ----------
        patch : np.ndarray
            Image patch, 2D or 3D, shape C(Z)YX.
        *args : Any
            Additional arguments, unused.
        **kwargs : Any
            Additional keyword arguments, unused.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            Masked patch, original patch, and mask.
        """
        masked = np.zeros_like(patch)
        mask = np.zeros_like(patch)
        if self.strategy == SupportedPixelManipulation.UNIFORM:
            # Iterate over the channels to apply manipulation separately
            for c in range(patch.shape[0]):
                masked[c, ...], mask[c, ...] = uniform_manipulate(
                    patch=patch[c, ...],
                    mask_pixel_percentage=self.masked_pixel_percentage,
                    subpatch_size=self.roi_size,
                    remove_center=self.remove_center,
                    struct_params=self.struct_mask,
                    rng=self.rng,
                )
        elif self.strategy == SupportedPixelManipulation.MEDIAN:
            # Iterate over the channels to apply manipulation separately
            for c in range(patch.shape[0]):
                masked[c, ...], mask[c, ...] = median_manipulate(
                    patch=patch[c, ...],
                    mask_pixel_percentage=self.masked_pixel_percentage,
                    subpatch_size=self.roi_size,
                    struct_params=self.struct_mask,
                    rng=self.rng,
                )
        else:
            raise ValueError(f"Unknown masking strategy ({self.strategy}).")

        # TODO: Output does not match other transforms, how to resolve?
        #     - Don't include in Compose and apply after if algorithm is N2V?
        #     - or just don't return patch? but then mask is in the target position
        # TODO why return patch?
        return masked, patch, mask



from pathlib import Path

import numpy as np
import pytest

from careamics.config import VAEAlgorithmConfig
from careamics.config.architectures import LVAEModel
from careamics.config.nm_model import (
    GaussianMixtureNMConfig,
    MultiChannelNMConfig,
)
from careamics.config.support import SupportedLoss


@pytest.mark.skip(
    reason="VAEAlgorithmConfig model is not currently serializable.\n"
    "The line `schema = VAEAlgorithmConfig.model_json_schema()` currently results "
    "in the following error:\n"
    "PydanticInvalidForJsonSchema: Cannot generate a JsonSchema for "
    "core_schema.IsInstanceSchema (<class 'torch.nn.modules.module.Module'>)"
)
def test_all_losses_are_supported():
    """Test that all losses defined in the Literal are supported."""
    # list of supported losses
    losses = list(SupportedLoss)

    # Algorithm json schema
    schema = VAEAlgorithmConfig.model_json_schema()

    # check that all losses are supported
    for loss in schema["properties"]["loss"]["enum"]:
        assert loss in losses


@pytest.mark.skip("Needs to be updated!")
def test_noise_model_usplit(minimum_algorithm_musplit):
    """Test that the noise model is correctly provided."""
    config = VAEAlgorithmConfig(**minimum_algorithm_musplit)
    assert config.noise_model is None


@pytest.mark.skip("Needs to be updated!")
def test_noise_model_denoisplit(tmp_path: Path, create_dummy_noise_model):
    """Test that the noise model is correctly provided."""
    # TODO this construct with the minimum_config dicts is increasingly annoying

    # Create a dummy noise model
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)

    # Instantiate the noise model
    gmm = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        path=tmp_path / "dummy_noise_model.npz",
        # all other params are default
    )
    config = VAEAlgorithmConfig(
        algorithm_type="vae",
        algorithm="denoisplit",
        loss="denoisplit",
        model=LVAEModel(architecture="LVAE"),
        noise_model=MultiChannelNMConfig(noise_models=[gmm]),
    )
    assert config.noise_model is not None


def test_no_noise_model_error_denoisplit(minimum_algorithm_denoisplit):
    """Test that the noise model is correctly provided."""
    minimum_algorithm_denoisplit["noise_model"] = None
    with pytest.raises(ValueError):
        VAEAlgorithmConfig(**minimum_algorithm_denoisplit)



"""
This script contains the utility functions for training the LVAE model.
These functions are mainly used in `train.py` script.
"""

import logging
import os
import pickle
import time
from copy import deepcopy
from datetime import datetime
from pathlib import Path

import ml_collections


def log_config(config: ml_collections.ConfigDict, cur_workdir: str) -> None:
    # Saving config file.
    with open(os.path.join(cur_workdir, "config.pkl"), "wb") as f:
        pickle.dump(config, f)
    print(f"Saved config to {cur_workdir}/config.pkl")


def set_logger(workdir: str) -> None:
    os.makedirs(workdir, exist_ok=True)
    fstream = open(os.path.join(workdir, "stdout.txt"), "w")
    handler = logging.StreamHandler(fstream)
    formatter = logging.Formatter(
        "%(levelname)s - %(filename)s - %(asctime)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger = logging.getLogger()
    logger.addHandler(handler)
    logger.setLevel("INFO")


def get_new_model_version(model_dir: str) -> str:
    """
    A model will have multiple runs. Each run will have a different version.
    """
    versions = []
    for version_dir in os.listdir(model_dir):
        try:
            versions.append(int(version_dir))
        except:
            print(
                f"Invalid subdirectory:{model_dir}/{version_dir}. Only integer versions are allowed"
            )
            exit()
    if len(versions) == 0:
        return "0"
    return f"{max(versions) + 1}"


def get_model_name(config: ml_collections.ConfigDict) -> str:
    return "LVAE_denoiSplit"


def get_workdir(
    config: ml_collections.ConfigDict,
    root_dir: str,
    use_max_version: bool,
    nested_call: int = 0,
):
    rel_path = datetime.now().strftime("%y%m")
    cur_workdir = os.path.join(root_dir, rel_path)
    Path(cur_workdir).mkdir(exist_ok=True)

    rel_path = os.path.join(rel_path, get_model_name(config))
    cur_workdir = os.path.join(root_dir, rel_path)
    Path(cur_workdir).mkdir(exist_ok=True)

    if use_max_version:
        # Used for debugging.
        version = int(get_new_model_version(cur_workdir))
        if version > 0:
            version = f"{version - 1}"

        rel_path = os.path.join(rel_path, str(version))
    else:
        rel_path = os.path.join(rel_path, get_new_model_version(cur_workdir))

    cur_workdir = os.path.join(root_dir, rel_path)
    try:
        Path(cur_workdir).mkdir(exist_ok=False)
    except FileExistsError:
        print(
            f"Workdir {cur_workdir} already exists. Probably because someother program also created the exact same directory. Trying to get a new version."
        )
        time.sleep(2.5)
        if nested_call > 10:
            raise ValueError(
                f"Cannot create a new directory. {cur_workdir} already exists."
            )

        return get_workdir(config, root_dir, use_max_version, nested_call + 1)

    return cur_workdir, rel_path


def get_mean_std_dict_for_model(config, train_dset):
    """
    Computes the mean and std for the model. This will be subsequently passed to the model.
    """
    mean_dict, std_dict = train_dset.get_mean_std()

    return deepcopy(mean_dict), deepcopy(std_dict)


class MetricMonitor:
    def __init__(self, metric):
        assert metric in ["val_loss", "val_psnr"]
        self.metric = metric

    def mode(self):
        if self.metric == "val_loss":
            return "min"
        elif self.metric == "val_psnr":
            return "max"
        else:
            raise ValueError(f"Invalid metric:{self.metric}")



from careamics.config.support import SupportedData
from careamics.file_io import get_write_func
from careamics.file_io.write import write_tiff


def test_get_write_tiff():
    assert get_write_func(SupportedData.TIFF) is write_tiff



"""Convenience functions to create configurations for training and inference."""

from typing import Any, Literal, Optional, Union

from .architectures import UNetModel
from .configuration_model import Configuration
from .data_model import DataConfig
from .fcn_algorithm_model import FCNAlgorithmConfig
from .support import (
    SupportedArchitecture,
    SupportedPixelManipulation,
    SupportedTransform,
)
from .training_model import TrainingConfig
from .transformations import (
    N2VManipulateModel,
    XYFlipModel,
    XYRandomRotate90Model,
)


def _list_augmentations(
    augmentations: Optional[list[Union[XYFlipModel, XYRandomRotate90Model]]],
) -> list[Union[XYFlipModel, XYRandomRotate90Model]]:
    """
    List the augmentations to apply.

    Parameters
    ----------
    augmentations : list of transforms, optional
        List of transforms to apply, either both or one of XYFlipModel and
        XYRandomRotate90Model.

    Returns
    -------
    list of transforms
        List of transforms to apply.

    Raises
    ------
    ValueError
        If the transforms are not XYFlipModel or XYRandomRotate90Model.
    ValueError
        If there are duplicate transforms.
    """
    if augmentations is None:
        transform_list: list[Union[XYFlipModel, XYRandomRotate90Model]] = [
            XYFlipModel(),
            XYRandomRotate90Model(),
        ]
    else:
        # throw error if not all transforms are pydantic models
        if not all(
            isinstance(t, XYFlipModel) or isinstance(t, XYRandomRotate90Model)
            for t in augmentations
        ):
            raise ValueError(
                "Accepted transforms are either XYFlipModel or "
                "XYRandomRotate90Model."
            )

        # check that there is no duplication
        aug_types = [t.__class__ for t in augmentations]
        if len(set(aug_types)) != len(aug_types):
            raise ValueError("Duplicate transforms are not allowed.")

        transform_list = augmentations

    return transform_list


def _create_unet_configuration(
    axes: str,
    n_channels_in: int,
    n_channels_out: int,
    independent_channels: bool,
    use_n2v2: bool,
    model_params: Optional[dict[str, Any]] = None,
) -> UNetModel:
    """
    Create a dictionary with the parameters of the UNet model.

    Parameters
    ----------
    axes : str
        Axes of the data.
    n_channels_in : int
        Number of input channels.
    n_channels_out : int
        Number of output channels.
    independent_channels : bool
        Whether to train all channels independently.
    use_n2v2 : bool
        Whether to use N2V2.
    model_params : dict
        UNetModel parameters.

    Returns
    -------
    UNetModel
        UNet model with the specified parameters.
    """
    if model_params is None:
        model_params = {}

    model_params["n2v2"] = use_n2v2
    model_params["conv_dims"] = 3 if "Z" in axes else 2
    model_params["in_channels"] = n_channels_in
    model_params["num_classes"] = n_channels_out
    model_params["independent_channels"] = independent_channels

    return UNetModel(
        architecture=SupportedArchitecture.UNET.value,
        **model_params,
    )


def _create_configuration(
    algorithm: Literal["n2v", "care", "n2n"],
    experiment_name: str,
    data_type: Literal["array", "tiff", "custom"],
    axes: str,
    patch_size: list[int],
    batch_size: int,
    num_epochs: int,
    augmentations: list[Union[XYFlipModel, XYRandomRotate90Model]],
    independent_channels: bool,
    loss: Literal["n2v", "mae", "mse"],
    n_channels_in: int,
    n_channels_out: int,
    logger: Literal["wandb", "tensorboard", "none"],
    use_n2v2: bool = False,
    model_params: Optional[dict] = None,
    dataloader_params: Optional[dict] = None,
) -> Configuration:
    """
    Create a configuration for training N2V, CARE or Noise2Noise.

    Parameters
    ----------
    algorithm : {"n2v", "care", "n2n"}
        Algorithm to use.
    experiment_name : str
        Name of the experiment.
    data_type : {"array", "tiff", "custom"}
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : list of int
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    num_epochs : int
        Number of epochs.
    augmentations : list of transforms
        List of transforms to apply, either both or one of XYFlipModel and
        XYRandomRotate90Model.
    independent_channels : bool
        Whether to train all channels independently.
    loss : {"n2v", "mae", "mse"}
        Loss function to use.
    n_channels_in : int
        Number of channels in.
    n_channels_out : int
        Number of channels out.
    logger : {"wandb", "tensorboard", "none"}
        Logger to use.
    use_n2v2 : bool, optional
        Whether to use N2V2, by default False.
    model_params : dict
        UNetModel parameters.
    dataloader_params : dict
        Parameters for the dataloader, see PyTorch notes, by default None.

    Returns
    -------
    Configuration
        Configuration for training N2V, CARE or Noise2Noise.
    """
    # model
    unet_model = _create_unet_configuration(
        axes=axes,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
        independent_channels=independent_channels,
        use_n2v2=use_n2v2,
        model_params=model_params,
    )

    # algorithm model
    algorithm_config = FCNAlgorithmConfig(
        algorithm=algorithm,
        loss=loss,
        model=unet_model,
    )

    # data model
    data = DataConfig(
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        transforms=augmentations,
        dataloader_params=dataloader_params,
    )

    # training model
    training = TrainingConfig(
        num_epochs=num_epochs,
        batch_size=batch_size,
        logger=None if logger == "none" else logger,
    )

    # create configuration
    configuration = Configuration(
        experiment_name=experiment_name,
        algorithm_config=algorithm_config,
        data_config=data,
        training_config=training,
    )

    return configuration


# TODO reconsider naming once we officially support LVAE approaches
def _create_supervised_configuration(
    algorithm: Literal["care", "n2n"],
    experiment_name: str,
    data_type: Literal["array", "tiff", "custom"],
    axes: str,
    patch_size: list[int],
    batch_size: int,
    num_epochs: int,
    augmentations: Optional[list[Union[XYFlipModel, XYRandomRotate90Model]]] = None,
    independent_channels: bool = True,
    loss: Literal["mae", "mse"] = "mae",
    n_channels_in: Optional[int] = None,
    n_channels_out: Optional[int] = None,
    logger: Literal["wandb", "tensorboard", "none"] = "none",
    model_params: Optional[dict] = None,
    dataloader_params: Optional[dict] = None,
) -> Configuration:
    """
    Create a configuration for training CARE or Noise2Noise.

    Parameters
    ----------
    algorithm : Literal["care", "n2n"]
        Algorithm to use.
    experiment_name : str
        Name of the experiment.
    data_type : Literal["array", "tiff", "custom"]
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : List[int]
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    num_epochs : int
        Number of epochs.
    augmentations : list of transforms, default=None
        List of transforms to apply, either both or one of XYFlipModel and
        XYRandomRotate90Model. By default, it applies both XYFlip (on X and Y)
        and XYRandomRotate90 (in XY) to the images.
    independent_channels : bool, optional
        Whether to train all channels independently, by default False.
    loss : Literal["mae", "mse"], optional
        Loss function to use, by default "mae".
    n_channels_in : int or None, default=None
        Number of channels in.
    n_channels_out : int or None, default=None
        Number of channels out.
    logger : Literal["wandb", "tensorboard", "none"], optional
        Logger to use, by default "none".
    model_params : dict, optional
        UNetModel parameters, by default {}.
    dataloader_params : dict, optional
        Parameters for the dataloader, see PyTorch notes, by default None.

    Returns
    -------
    Configuration
        Configuration for training CARE or Noise2Noise.

    Raises
    ------
    ValueError
        If the number of channels is not specified when using channels.
    ValueError
        If the number of channels is specified but "C" is not in the axes.
    """
    # if there are channels, we need to specify their number
    if "C" in axes and n_channels_in is None:
        raise ValueError("Number of channels in must be specified when using channels ")
    elif "C" not in axes and (n_channels_in is not None and n_channels_in > 1):
        raise ValueError(
            f"C is not present in the axes, but number of channels is specified "
            f"(got {n_channels_in} channels)."
        )

    if n_channels_in is None:
        n_channels_in = 1

    if n_channels_out is None:
        n_channels_out = n_channels_in

    # augmentations
    transform_list = _list_augmentations(augmentations)

    return _create_configuration(
        algorithm=algorithm,
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        augmentations=transform_list,
        independent_channels=independent_channels,
        loss=loss,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
        logger=logger,
        model_params=model_params,
        dataloader_params=dataloader_params,
    )


def create_care_configuration(
    experiment_name: str,
    data_type: Literal["array", "tiff", "custom"],
    axes: str,
    patch_size: list[int],
    batch_size: int,
    num_epochs: int,
    augmentations: Optional[list[Union[XYFlipModel, XYRandomRotate90Model]]] = None,
    independent_channels: bool = True,
    loss: Literal["mae", "mse"] = "mae",
    n_channels_in: Optional[int] = None,
    n_channels_out: Optional[int] = None,
    logger: Literal["wandb", "tensorboard", "none"] = "none",
    model_params: Optional[dict] = None,
    dataloader_params: Optional[dict] = None,
) -> Configuration:
    """
    Create a configuration for training CARE.

    If "Z" is present in `axes`, then `path_size` must be a list of length 3, otherwise
    2.

    If "C" is present in `axes`, then you need to set `n_channels_in` to the number of
    channels. Likewise, if you set the number of channels, then "C" must be present in
    `axes`.

    To set the number of output channels, use the `n_channels_out` parameter. If it is
    not specified, it will be assumed to be equal to `n_channels_in`.

    By default, all channels are trained together. To train all channels independently,
    set `independent_channels` to True.

    By setting `augmentations` to `None`, the default transformations (flip in X and Y,
    rotations by 90 degrees in the XY plane) are applied. Rather than the default
    transforms, a list of transforms can be passed to the `augmentations` parameter. To
    disable the transforms, simply pass an empty list.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    data_type : Literal["array", "tiff", "custom"]
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : List[int]
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    num_epochs : int
        Number of epochs.
    augmentations : list of transforms, default=None
        List of transforms to apply, either both or one of XYFlipModel and
        XYRandomRotate90Model. By default, it applies both XYFlip (on X and Y)
        and XYRandomRotate90 (in XY) to the images.
    independent_channels : bool, optional
        Whether to train all channels independently, by default False.
    loss : Literal["mae", "mse"], default="mae"
        Loss function to use.
    n_channels_in : int or None, default=None
        Number of channels in.
    n_channels_out : int or None, default=None
        Number of channels out.
    logger : Literal["wandb", "tensorboard", "none"], default="none"
        Logger to use.
    model_params : dict, default=None
        UNetModel parameters.
    dataloader_params : dict, optional
        Parameters for the dataloader, see PyTorch notes, by default None.

    Returns
    -------
    Configuration
        Configuration for training CARE.

    Examples
    --------
    Minimum example:
    >>> config = create_care_configuration(
    ...     experiment_name="care_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100
    ... )

    To disable transforms, simply set `augmentations` to an empty list:
    >>> config = create_care_configuration(
    ...     experiment_name="care_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     augmentations=[]
    ... )

    A list of transforms can be passed to the `augmentations` parameter to replace the
    default augmentations:
    >>> from careamics.config.transformations import XYFlipModel
    >>> config = create_care_configuration(
    ...     experiment_name="care_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     augmentations=[
    ...         # No rotation and only Y flipping
    ...         XYFlipModel(flip_x = False, flip_y = True)
    ...     ]
    ... )

    If you are training multiple channels they will be trained independently by default,
    you simply need to specify the number of channels input (and optionally, the number
    of channels output):
    >>> config = create_care_configuration(
    ...     experiment_name="care_experiment",
    ...     data_type="array",
    ...     axes="YXC", # channels must be in the axes
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     n_channels_in=3, # number of input channels
    ...     n_channels_out=1 # if applicable
    ... )

    If instead you want to train multiple channels together, you need to turn off the
    `independent_channels` parameter:
    >>> config = create_care_configuration(
    ...     experiment_name="care_experiment",
    ...     data_type="array",
    ...     axes="YXC", # channels must be in the axes
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     independent_channels=False,
    ...     n_channels_in=3,
    ...     n_channels_out=1 # if applicable
    ... )
    """
    return _create_supervised_configuration(
        algorithm="care",
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        augmentations=augmentations,
        independent_channels=independent_channels,
        loss=loss,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
        logger=logger,
        model_params=model_params,
        dataloader_params=dataloader_params,
    )


def create_n2n_configuration(
    experiment_name: str,
    data_type: Literal["array", "tiff", "custom"],
    axes: str,
    patch_size: list[int],
    batch_size: int,
    num_epochs: int,
    augmentations: Optional[list[Union[XYFlipModel, XYRandomRotate90Model]]] = None,
    independent_channels: bool = True,
    loss: Literal["mae", "mse"] = "mae",
    n_channels_in: Optional[int] = None,
    n_channels_out: Optional[int] = None,
    logger: Literal["wandb", "tensorboard", "none"] = "none",
    model_params: Optional[dict] = None,
    dataloader_params: Optional[dict] = None,
) -> Configuration:
    """
    Create a configuration for training Noise2Noise.

    If "Z" is present in `axes`, then `path_size` must be a list of length 3, otherwise
    2.

    If "C" is present in `axes`, then you need to set `n_channels_in` to the number of
    channels. Likewise, if you set the number of channels, then "C" must be present in
    `axes`.

    To set the number of output channels, use the `n_channels_out` parameter. If it is
    not specified, it will be assumed to be equal to `n_channels_in`.

    By default, all channels are trained together. To train all channels independently,
    set `independent_channels` to True.

    By setting `augmentations` to `None`, the default transformations (flip in X and Y,
    rotations by 90 degrees in the XY plane) are applied. Rather than the default
    transforms, a list of transforms can be passed to the `augmentations` parameter. To
    disable the transforms, simply pass an empty list.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    data_type : Literal["array", "tiff", "custom"]
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : List[int]
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    num_epochs : int
        Number of epochs.
    augmentations : list of transforms, default=None
        List of transforms to apply, either both or one of XYFlipModel and
        XYRandomRotate90Model. By default, it applies both XYFlip (on X and Y)
        and XYRandomRotate90 (in XY) to the images.
    independent_channels : bool, optional
        Whether to train all channels independently, by default False.
    loss : Literal["mae", "mse"], optional
        Loss function to use, by default "mae".
    n_channels_in : int or None, default=None
        Number of channels in.
    n_channels_out : int or None, default=None
        Number of channels out.
    logger : Literal["wandb", "tensorboard", "none"], optional
        Logger to use, by default "none".
    model_params : dict, optional
        UNetModel parameters, by default {}.
    dataloader_params : dict, optional
        Parameters for the dataloader, see PyTorch notes, by default None.

    Returns
    -------
    Configuration
        Configuration for training Noise2Noise.

    Examples
    --------
    Minimum example:
    >>> config = create_n2n_configuration(
    ...     experiment_name="n2n_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100
    ... )

    To disable transforms, simply set `augmentations` to an empty list:
    >>> config = create_n2n_configuration(
    ...     experiment_name="n2n_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     augmentations=[]
    ... )

    A list of transforms can be passed to the `augmentations` parameter to replace the
    default augmentations:
    >>> from careamics.config.transformations import XYFlipModel
    >>> config = create_n2n_configuration(
    ...     experiment_name="n2n_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     augmentations=[
    ...         # No rotation and only Y flipping
    ...         XYFlipModel(flip_x = False, flip_y = True)
    ...     ]
    ... )

    If you are training multiple channels they will be trained independently by default,
    you simply need to specify the number of channels input (and optionally, the number
    of channels output):
    >>> config = create_n2n_configuration(
    ...     experiment_name="n2n_experiment",
    ...     data_type="array",
    ...     axes="YXC", # channels must be in the axes
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     n_channels_in=3, # number of input channels
    ...     n_channels_out=1 # if applicable
    ... )

    If instead you want to train multiple channels together, you need to turn off the
    `independent_channels` parameter:
    >>> config = create_n2n_configuration(
    ...     experiment_name="n2n_experiment",
    ...     data_type="array",
    ...     axes="YXC", # channels must be in the axes
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     independent_channels=False,
    ...     n_channels_in=3,
    ...     n_channels_out=1 # if applicable
    ... )
    """
    return _create_supervised_configuration(
        algorithm="n2n",
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        augmentations=augmentations,
        independent_channels=independent_channels,
        loss=loss,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
        logger=logger,
        model_params=model_params,
        dataloader_params=dataloader_params,
    )


def create_n2v_configuration(
    experiment_name: str,
    data_type: Literal["array", "tiff", "custom"],
    axes: str,
    patch_size: list[int],
    batch_size: int,
    num_epochs: int,
    augmentations: Optional[list[Union[XYFlipModel, XYRandomRotate90Model]]] = None,
    independent_channels: bool = True,
    use_n2v2: bool = False,
    n_channels: Optional[int] = None,
    roi_size: int = 11,
    masked_pixel_percentage: float = 0.2,
    struct_n2v_axis: Literal["horizontal", "vertical", "none"] = "none",
    struct_n2v_span: int = 5,
    logger: Literal["wandb", "tensorboard", "none"] = "none",
    model_params: Optional[dict] = None,
    dataloader_params: Optional[dict] = None,
) -> Configuration:
    """
    Create a configuration for training Noise2Void.

    N2V uses a UNet model to denoise images in a self-supervised manner. To use its
    variants structN2V and N2V2, set the `struct_n2v_axis` and `struct_n2v_span`
    (structN2V) parameters, or set `use_n2v2` to True (N2V2).

    N2V2 modifies the UNet architecture by adding blur pool layers and removes the skip
    connections, thus removing checkboard artefacts. StructN2V is used when vertical
    or horizontal correlations are present in the noise; it applies an additional mask
    to the manipulated pixel neighbors.

    If "Z" is present in `axes`, then `path_size` must be a list of length 3, otherwise
    2.

    If "C" is present in `axes`, then you need to set `n_channels` to the number of
    channels.

    By default, all channels are trained independently. To train all channels together,
    set `independent_channels` to False.

    By default, the transformations applied are a random flip along X or Y, and a random
    90 degrees rotation in the XY plane. Normalization is always applied, as well as the
    N2V manipulation.

    By setting `augmentations` to `None`, the default transformations (flip in X and Y,
    rotations by 90 degrees in the XY plane) are applied. Rather than the default
    transforms, a list of transforms can be passed to the `augmentations` parameter. To
    disable the transforms, simply pass an empty list.

    The `roi_size` parameter specifies the size of the area around each pixel that will
    be manipulated by N2V. The `masked_pixel_percentage` parameter specifies how many
    pixels per patch will be manipulated.

    The parameters of the UNet can be specified in the `model_params` (passed as a
    parameter-value dictionary). Note that `use_n2v2` and 'n_channels' override the
    corresponding parameters passed in `model_params`.

    If you pass "horizontal" or "vertical" to `struct_n2v_axis`, then structN2V mask
    will be applied to each manipulated pixel.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    data_type : Literal["array", "tiff", "custom"]
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : List[int]
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    num_epochs : int
        Number of epochs.
    augmentations : list of transforms, default=None
        List of transforms to apply, either both or one of XYFlipModel and
        XYRandomRotate90Model. By default, it applies both XYFlip (on X and Y)
        and XYRandomRotate90 (in XY) to the images.
    independent_channels : bool, optional
        Whether to train all channels together, by default True.
    use_n2v2 : bool, optional
        Whether to use N2V2, by default False.
    n_channels : int or None, default=None
        Number of channels (in and out).
    roi_size : int, optional
        N2V pixel manipulation area, by default 11.
    masked_pixel_percentage : float, optional
        Percentage of pixels masked in each patch, by default 0.2.
    struct_n2v_axis : Literal["horizontal", "vertical", "none"], optional
        Axis along which to apply structN2V mask, by default "none".
    struct_n2v_span : int, optional
        Span of the structN2V mask, by default 5.
    logger : Literal["wandb", "tensorboard", "none"], optional
        Logger to use, by default "none".
    model_params : dict, optional
        UNetModel parameters, by default None.
    dataloader_params : dict, optional
        Parameters for the dataloader, see PyTorch notes, by default None.

    Returns
    -------
    Configuration
        Configuration for training N2V.

    Examples
    --------
    Minimum example:
    >>> config = create_n2v_configuration(
    ...     experiment_name="n2v_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100
    ... )

    To disable transforms, simply set `augmentations` to an empty list:
    >>> config = create_n2v_configuration(
    ...     experiment_name="n2v_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     augmentations=[]
    ... )

    A list of transforms can be passed to the `augmentations` parameter:
    >>> from careamics.config.transformations import XYFlipModel
    >>> config = create_n2v_configuration(
    ...     experiment_name="n2v_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     augmentations=[
    ...         # No rotation and only Y flipping
    ...         XYFlipModel(flip_x = False, flip_y = True)
    ...     ]
    ... )

    To use N2V2, simply pass the `use_n2v2` parameter:
    >>> config = create_n2v_configuration(
    ...     experiment_name="n2v2_experiment",
    ...     data_type="tiff",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     use_n2v2=True
    ... )

    For structN2V, there are two parameters to set, `struct_n2v_axis` and
    `struct_n2v_span`:
    >>> config = create_n2v_configuration(
    ...     experiment_name="structn2v_experiment",
    ...     data_type="tiff",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     struct_n2v_axis="horizontal",
    ...     struct_n2v_span=7
    ... )

    If you are training multiple channels they will be trained independently by default,
    you simply need to specify the number of channels:
    >>> config = create_n2v_configuration(
    ...     experiment_name="n2v_experiment",
    ...     data_type="array",
    ...     axes="YXC",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     n_channels=3
    ... )

    If instead you want to train multiple channels together, you need to turn off the
    `independent_channels` parameter:
    >>> config = create_n2v_configuration(
    ...     experiment_name="n2v_experiment",
    ...     data_type="array",
    ...     axes="YXC",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100,
    ...     independent_channels=False,
    ...     n_channels=3
    ... )
    """
    # if there are channels, we need to specify their number
    if "C" in axes and n_channels is None:
        raise ValueError("Number of channels must be specified when using channels.")
    elif "C" not in axes and (n_channels is not None and n_channels > 1):
        raise ValueError(
            f"C is not present in the axes, but number of channels is specified "
            f"(got {n_channels} channel)."
        )

    if n_channels is None:
        n_channels = 1

    # augmentations
    transform_list = _list_augmentations(augmentations)

    # create the N2VManipulate transform using the supplied parameters
    n2v_transform = N2VManipulateModel(
        name=SupportedTransform.N2V_MANIPULATE.value,
        strategy=(
            SupportedPixelManipulation.MEDIAN.value
            if use_n2v2
            else SupportedPixelManipulation.UNIFORM.value
        ),
        roi_size=roi_size,
        masked_pixel_percentage=masked_pixel_percentage,
        struct_mask_axis=struct_n2v_axis,
        struct_mask_span=struct_n2v_span,
    )
    transform_list.append(n2v_transform)

    return _create_configuration(
        algorithm="n2v",
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        augmentations=transform_list,
        independent_channels=independent_channels,
        loss="n2v",
        use_n2v2=use_n2v2,
        n_channels_in=n_channels,
        n_channels_out=n_channels,
        logger=logger,
        model_params=model_params,
        dataloader_params=dataloader_params,
    )



import pytest

from careamics.config import (
    create_care_configuration,
    create_n2n_configuration,
    create_n2v_configuration,
)
from careamics.config.configuration_factory import (
    _create_configuration,
    _create_supervised_configuration,
    _create_unet_configuration,
    _list_augmentations,
)
from careamics.config.support import (
    SupportedPixelManipulation,
    SupportedStructAxis,
    SupportedTransform,
)
from careamics.config.transformations import (
    N2VManipulateModel,
    XYFlipModel,
    XYRandomRotate90Model,
)


def test_list_aug_default():
    """Test that the default augmentations are present."""
    list_aug = _list_augmentations(augmentations=None)

    assert len(list_aug) == 2
    assert list_aug[0].name == SupportedTransform.XY_FLIP.value
    assert list_aug[1].name == SupportedTransform.XY_RANDOM_ROTATE90.value


def test_list_aug_no_aug():
    """Test that disabling augmentation results in empty transform list."""
    list_aug = _list_augmentations(augmentations=[])
    assert len(list_aug) == 0


def test_list_aug_error_duplicate_transforms():
    """Test that an error is raised when there are duplicate transforms."""
    with pytest.raises(ValueError):
        _list_augmentations(
            augmentations=[XYFlipModel(), XYRandomRotate90Model(), XYFlipModel()],
        )


def test_list_aug_error_wrong_transform():
    """Test that an error is raised when the wrong transform is passed."""
    with pytest.raises(ValueError):
        _list_augmentations(
            augmentations=[XYFlipModel(), N2VManipulateModel()],
        )


def test_supervised_configuration_passing_transforms():
    """Test that transforms can be passed to the configuration."""
    config = _create_supervised_configuration(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        augmentations=[XYFlipModel()],
    )
    assert len(config.data_config.transforms) == 1
    assert config.data_config.transforms[0].name == SupportedTransform.XY_FLIP.value


def test_model_creation():
    """Test that the correct parameters are passed to the model."""
    model_kwargs = {
        "depth": 4,
        "conv_dims": 2,
        "n2v2": False,
        "in_channels": 2,
        "num_classes": 5,
        "independent_channels": False,
    }

    # choose different parameters
    axes = "XYZ"
    conv_dims = 3
    in_channels = 3
    num_classes = 4
    independent_channels = True
    use_n2v2 = True

    model = _create_unet_configuration(
        axes=axes,
        n_channels_in=in_channels,
        n_channels_out=num_classes,
        independent_channels=independent_channels,
        use_n2v2=use_n2v2,
        model_params=model_kwargs,
    )

    assert model.depth == model_kwargs["depth"]
    assert model.conv_dims == conv_dims
    assert model.n2v2 == use_n2v2
    assert model.in_channels == in_channels
    assert model.num_classes == num_classes
    assert model.independent_channels == independent_channels


def test_create_configuration():
    """Test that the methods correctly passes all parameters."""
    algorithm = "care"
    experiment_name = "test"
    data_type = "tiff"
    axes = "CYX"
    patch_size = [128, 128]
    batch_size = 8
    num_epochs = 100
    transform_list = [XYFlipModel(), XYRandomRotate90Model()]
    independent_channels = False
    loss = "mse"
    n_channels_in = 2
    n_channels_out = 3
    logger = "tensorboard"
    model_params = {
        "depth": 5,
    }
    dataloader_params = {
        "num_workers": 4,
    }

    # instantiate config
    config = _create_configuration(
        algorithm=algorithm,
        experiment_name=experiment_name,
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        num_epochs=num_epochs,
        augmentations=transform_list,
        independent_channels=independent_channels,
        loss=loss,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
        logger=logger,
        model_params=model_params,
        dataloader_params=dataloader_params,
    )

    assert config.algorithm_config.algorithm == algorithm
    assert config.experiment_name == experiment_name
    assert config.data_config.data_type == data_type
    assert config.data_config.axes == axes
    assert config.data_config.patch_size == patch_size
    assert config.data_config.batch_size == batch_size
    assert config.training_config.num_epochs == num_epochs
    assert config.data_config.transforms == transform_list
    assert config.algorithm_config.model.independent_channels == independent_channels
    assert config.algorithm_config.loss == loss
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out
    assert config.training_config.logger == logger
    assert config.algorithm_config.model.depth == model_params["depth"]
    assert config.data_config.dataloader_params == dataloader_params


def test_supervised_configuration_error_with_channel_axes():
    """Test that an error is raised if channels are in axes, but the input channel
    number is not specified."""
    with pytest.raises(ValueError):
        _create_supervised_configuration(
            algorithm="n2n",
            experiment_name="test",
            data_type="tiff",
            axes="CYX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=100,
        )


def test_supervised_configuration_singleton_channel():
    """Test that no error is raised if channels are in axes, and the input channel is
    1."""
    _create_supervised_configuration(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=1,
    )


def test_supervised_configuration_no_channel():
    """Test that no error is raised without channel and number of inputs."""
    _create_supervised_configuration(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
    )


def test_supervised_configuration_error_without_channel_axes():
    """Test that an error is raised if channels are not in axes, but the input channel
    number is specified and greater than 1."""
    with pytest.raises(ValueError):
        _create_supervised_configuration(
            algorithm="n2n",
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=100,
            n_channels_in=2,
        )


def test_supervised_configuration_channels():
    """Test that no error is raised if channels are in axes and the input channel
    are specified."""
    _create_supervised_configuration(
        algorithm="n2n",
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=4,
    )


def test_n2n_configuration():
    """Test that N2N configuration can be created."""
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
    )
    assert config.algorithm_config.algorithm == "n2n"


def test_n2n_configuration_n_channels():
    """Test the behaviour of the number of channels in and out."""
    n_channels_in = 4
    n_channels_out = 5

    # n_channels_out not specified
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=n_channels_in,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_in

    # specify n_channels_out
    config = create_n2n_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out


def test_care_configuration():
    """Test that CARE configuration can be created."""
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
    )
    assert config.algorithm_config.algorithm == "care"


def test_care_configuration_n_channels():
    """Test the behaviour of the number of channels in and out."""
    n_channels_in = 4
    n_channels_out = 5

    # n_channels_out not specified
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=n_channels_in,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_in

    # specify n_channels_out
    config = create_care_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="CYX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        n_channels_in=n_channels_in,
        n_channels_out=n_channels_out,
    )
    assert config.algorithm_config.model.in_channels == n_channels_in
    assert config.algorithm_config.model.num_classes == n_channels_out


def test_n2v_configuration():
    """Test that N2V configuration can be created."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
    )
    assert config.algorithm_config.algorithm == "n2v"
    assert config.algorithm_config.loss == "n2v"
    assert (
        config.data_config.transforms[-1].name
        == SupportedTransform.N2V_MANIPULATE.value
    )


def test_n2v_configuration_default_transforms():
    """Test the default n2v transforms."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
    )
    assert len(config.data_config.transforms) == 3
    assert config.data_config.transforms[0].name == SupportedTransform.XY_FLIP.value
    assert (
        config.data_config.transforms[1].name
        == SupportedTransform.XY_RANDOM_ROTATE90.value
    )
    assert (
        config.data_config.transforms[2].name == SupportedTransform.N2V_MANIPULATE.value
    )


def test_n2v_configuration_no_aug():
    """Test the default n2v transforms."""
    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        augmentations=[],
    )
    assert len(config.data_config.transforms) == 1
    assert (
        config.data_config.transforms[-1].name
        == SupportedTransform.N2V_MANIPULATE.value
    )


def test_n2v_configuration_n2v2_structn2v():
    """Test that N2V manipulate is ignored when explicitely passed."""
    use_n2v2 = True
    roi_size = 15
    masked_pixel_percentage = 0.5
    struct_mask_axis = SupportedStructAxis.HORIZONTAL.value
    struct_n2v_span = 15

    config = create_n2v_configuration(
        experiment_name="test",
        data_type="tiff",
        axes="YX",
        patch_size=[64, 64],
        batch_size=8,
        num_epochs=100,
        use_n2v2=use_n2v2,  # median strategy
        roi_size=roi_size,
        masked_pixel_percentage=masked_pixel_percentage,
        struct_n2v_axis=struct_mask_axis,
        struct_n2v_span=struct_n2v_span,
    )
    assert len(config.data_config.transforms) == 3
    assert (
        config.data_config.transforms[-1].name
        == SupportedTransform.N2V_MANIPULATE.value
    )
    assert (
        config.data_config.transforms[-1].name
        == SupportedTransform.N2V_MANIPULATE.value
    )
    assert (
        config.data_config.transforms[-1].strategy
        == SupportedPixelManipulation.MEDIAN.value
    )
    assert config.data_config.transforms[-1].roi_size == roi_size
    assert (
        config.data_config.transforms[-1].masked_pixel_percentage
        == masked_pixel_percentage
    )
    assert config.data_config.transforms[-1].struct_mask_axis == struct_mask_axis
    assert config.data_config.transforms[-1].struct_mask_span == struct_n2v_span



"""In-memory dataset module."""

from __future__ import annotations

import copy
from pathlib import Path
from typing import Any, Callable, Optional, Union

import numpy as np
from torch.utils.data import Dataset

from careamics.file_io.read import read_tiff
from careamics.transforms import Compose

from ..config import DataConfig
from ..config.transformations import NormalizeModel
from ..utils.logging import get_logger
from .patching.patching import (
    PatchedOutput,
    Stats,
    prepare_patches_supervised,
    prepare_patches_supervised_array,
    prepare_patches_unsupervised,
    prepare_patches_unsupervised_array,
)

logger = get_logger(__name__)


class InMemoryDataset(Dataset):
    """Dataset storing data in memory and allowing generating patches from it.

    Parameters
    ----------
    data_config : CAREamics DataConfig
        (see careamics.config.data_model.DataConfig)
        Data configuration.
    inputs : numpy.ndarray or list[pathlib.Path]
        Input data.
    input_target : numpy.ndarray or list[pathlib.Path], optional
        Target data, by default None.
    read_source_func : Callable, optional
        Read source function for custom types, by default read_tiff.
    **kwargs : Any
        Additional keyword arguments, unused.
    """

    def __init__(
        self,
        data_config: DataConfig,
        inputs: Union[np.ndarray, list[Path]],
        input_target: Optional[Union[np.ndarray, list[Path]]] = None,
        read_source_func: Callable = read_tiff,
        **kwargs: Any,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        data_config : DataConfig
            Data configuration.
        inputs : numpy.ndarray or list[pathlib.Path]
            Input data.
        input_target : numpy.ndarray or list[pathlib.Path], optional
            Target data, by default None.
        read_source_func : Callable, optional
            Read source function for custom types, by default read_tiff.
        **kwargs : Any
            Additional keyword arguments, unused.
        """
        self.data_config = data_config
        self.inputs = inputs
        self.input_targets = input_target
        self.axes = self.data_config.axes
        self.patch_size = self.data_config.patch_size

        # read function
        self.read_source_func = read_source_func

        # generate patches
        supervised = self.input_targets is not None
        patches_data = self._prepare_patches(supervised)

        # unpack the dataclass
        self.data = patches_data.patches
        self.data_targets = patches_data.targets

        # set image statistics
        if self.data_config.image_means is None:
            self.image_stats = patches_data.image_stats
            logger.info(
                f"Computed dataset mean: {self.image_stats.means}, "
                f"std: {self.image_stats.stds}"
            )
        else:
            self.image_stats = Stats(
                self.data_config.image_means, self.data_config.image_stds
            )

        # set target statistics
        if self.data_config.target_means is None:
            self.target_stats = patches_data.target_stats
        else:
            self.target_stats = Stats(
                self.data_config.target_means, self.data_config.target_stds
            )

        # update mean and std in configuration
        # the object is mutable and should then be recorded in the CAREamist obj
        self.data_config.set_means_and_stds(
            image_means=self.image_stats.means,
            image_stds=self.image_stats.stds,
            target_means=self.target_stats.means,
            target_stds=self.target_stats.stds,
        )
        # get transforms
        self.patch_transform = Compose(
            transform_list=[
                NormalizeModel(
                    image_means=self.image_stats.means,
                    image_stds=self.image_stats.stds,
                    target_means=self.target_stats.means,
                    target_stds=self.target_stats.stds,
                )
            ]
            + self.data_config.transforms,
        )

    def _prepare_patches(self, supervised: bool) -> PatchedOutput:
        """
        Iterate over data source and create an array of patches.

        Parameters
        ----------
        supervised : bool
            Whether the dataset is supervised or not.

        Returns
        -------
        numpy.ndarray
            Array of patches.
        """
        if supervised:
            if isinstance(self.inputs, np.ndarray) and isinstance(
                self.input_targets, np.ndarray
            ):
                return prepare_patches_supervised_array(
                    self.inputs,
                    self.axes,
                    self.input_targets,
                    self.patch_size,
                )
            elif isinstance(self.inputs, list) and isinstance(self.input_targets, list):
                return prepare_patches_supervised(
                    self.inputs,
                    self.input_targets,
                    self.axes,
                    self.patch_size,
                    self.read_source_func,
                )
            else:
                raise ValueError(
                    f"Data and target must be of the same type, either both numpy "
                    f"arrays or both lists of paths, got {type(self.inputs)} (data) "
                    f"and {type(self.input_targets)} (target)."
                )
        else:
            if isinstance(self.inputs, np.ndarray):
                return prepare_patches_unsupervised_array(
                    self.inputs,
                    self.axes,
                    self.patch_size,
                )
            else:
                return prepare_patches_unsupervised(
                    self.inputs,
                    self.axes,
                    self.patch_size,
                    self.read_source_func,
                )

    def __len__(self) -> int:
        """
        Return the length of the dataset.

        Returns
        -------
        int
            Length of the dataset.
        """
        return self.data.shape[0]

    def __getitem__(self, index: int) -> tuple[np.ndarray, ...]:
        """
        Return the patch corresponding to the provided index.

        Parameters
        ----------
        index : int
            Index of the patch to return.

        Returns
        -------
        tuple of numpy.ndarray
            Patch.

        Raises
        ------
        ValueError
            If dataset mean and std are not set.
        """
        patch = self.data[index]

        # if there is a target
        if self.data_targets is not None:
            # get target
            target = self.data_targets[index]

            return self.patch_transform(patch=patch, target=target)

        elif self.data_config.has_n2v_manipulate():  # TODO not compatible with HDN
            return self.patch_transform(patch=patch)
        else:
            raise ValueError(
                "Something went wrong! No target provided (not supervised training) "
                "and no N2V manipulation (no N2V training)."
            )

    def get_data_statistics(self) -> tuple[list[float], list[float]]:
        """Return training data statistics.

        This does not return the target data statistics, only those of the input.

        Returns
        -------
        tuple of list of floats
            Means and standard deviations across channels of the training data.
        """
        return self.image_stats.get_statistics()

    def split_dataset(
        self,
        percentage: float = 0.1,
        minimum_patches: int = 1,
    ) -> InMemoryDataset:
        """Split a new dataset away from the current one.

        This method is used to extract random validation patches from the dataset.

        Parameters
        ----------
        percentage : float, optional
            Percentage of patches to extract, by default 0.1.
        minimum_patches : int, optional
            Minimum number of patches to extract, by default 5.

        Returns
        -------
        CAREamics InMemoryDataset
            New dataset with the extracted patches.

        Raises
        ------
        ValueError
            If `percentage` is not between 0 and 1.
        ValueError
            If `minimum_number` is not between 1 and the number of patches.
        """
        if percentage < 0 or percentage > 1:
            raise ValueError(f"Percentage must be between 0 and 1, got {percentage}.")

        if minimum_patches < 1 or minimum_patches > len(self):
            raise ValueError(
                f"Minimum number of patches must be between 1 and "
                f"{len(self)} (number of patches), got "
                f"{minimum_patches}. Adjust the patch size or the minimum number of "
                f"patches."
            )

        total_patches = len(self)

        # number of patches to extract (either percentage rounded or minimum number)
        n_patches = max(round(total_patches * percentage), minimum_patches)

        # get random indices
        indices = np.random.choice(total_patches, n_patches, replace=False)

        # extract patches
        val_patches = self.data[indices]

        # remove patches from self.patch
        self.data = np.delete(self.data, indices, axis=0)

        # same for targets
        if self.data_targets is not None:
            val_targets = self.data_targets[indices]
            self.data_targets = np.delete(self.data_targets, indices, axis=0)

        # clone the dataset
        dataset = copy.deepcopy(self)

        # reassign patches
        dataset.data = val_patches

        # reassign targets
        if self.data_targets is not None:
            dataset.data_targets = val_targets

        return dataset



import os

import numpy as np
import pytest
import tifffile

from careamics.config import DataConfig
from careamics.config.support import SupportedData
from careamics.dataset import PathIterableDataset
from careamics.file_io.read import read_tiff


@pytest.mark.parametrize(
    "shape, axes, patch_size",
    [
        ((32, 32), "YX", (8, 8)),
        ((2, 32, 32), "CYX", (8, 8)),
        ((32, 32, 32), "ZYX", (8, 8, 8)),
    ],
)
def test_number_of_files(tmp_path, ordered_array, shape, axes, patch_size):
    """Test number of files in PathIterableDataset."""
    # create array
    array_size = 32
    n_files = 3
    factor = 3 if axes == "ZYX" else 2
    array = ordered_array(shape)

    # save three files
    files = []
    for i in range(n_files):
        file = tmp_path / f"array{i}.tif"
        tifffile.imwrite(file, array)
        files.append(file)

    # create config
    config_dict = {
        "data_type": SupportedData.TIFF.value,
        "patch_size": patch_size,
        "axes": axes,
    }
    config = DataConfig(**config_dict)

    # create dataset
    dataset = PathIterableDataset(
        data_config=config, src_files=files, read_source_func=read_tiff
    )

    # check number of files
    assert dataset.data_files == files

    # iterate over dataset
    patches = list(dataset)
    assert len(patches) == n_files * (array_size / patch_size[0]) ** factor


def test_read_function(tmp_path, ordered_array):
    """Test reading files in PathIterableDataset using a custom read function."""

    # read function for .npy files
    def read_npy(file_path, *args, **kwargs):
        return np.load(file_path)

    array_size = 32
    patch_size = 8
    n_files = 3
    patch_sizes = [patch_size] * 2

    # create array
    array: np.ndarray = ordered_array((n_files, array_size, array_size))

    # save each plane in a single .npy file
    files = []
    for i in range(array.shape[0]):
        file_path = tmp_path / f"array{i}.npy"
        np.save(file_path, array[i])
        files.append(file_path)

    # create config
    config_dict = {
        "data_type": SupportedData.CUSTOM.value,
        "patch_size": patch_sizes,
        "axes": "YX",
    }
    config = DataConfig(**config_dict)

    # create dataset
    dataset = PathIterableDataset(
        data_config=config,
        src_files=files,
        read_source_func=read_npy,
    )
    assert dataset.data_files == files

    # iterate over dataset
    patches = list(dataset)
    assert len(patches) == n_files * (array_size / patch_size) ** 2


@pytest.mark.parametrize("percentage", [0.1, 0.6])
def test_extracting_val_files(tmp_path, ordered_array, percentage):
    """Test extracting a validation set patches from PathIterableDataset."""
    # create array
    array = ordered_array((20, 20))

    # save array to 25 files
    files = []
    for i in range(25):
        file_path = tmp_path / f"array{i}.tif"
        tifffile.imwrite(file_path, array)
        files.append(file_path)

    # create config
    config_dict = {
        "data_type": SupportedData.TIFF.value,
        "patch_size": [8, 8],
        "axes": "YX",
    }
    config = DataConfig(**config_dict)

    # create dataset
    dataset = PathIterableDataset(
        data_config=config, src_files=files, read_source_func=read_tiff
    )

    # compute number of patches
    total_n_files = dataset.get_number_of_files()
    minimum_files = 5
    n_files = max(round(percentage * total_n_files), minimum_files)

    # extract datset
    valset = dataset.split_dataset(percentage, minimum_files)

    # check number of patches
    assert valset.get_number_of_files() == n_files
    assert dataset.get_number_of_files() == total_n_files - n_files

    # check that none of the validation files are in the original dataset
    assert set(valset.data_files).isdisjoint(set(dataset.data_files))


@pytest.mark.parametrize(
    "shape, axes, patch_size",
    [
        ((32, 32), "YX", (8, 8)),
        ((2, 32, 32), "CYX", (8, 8)),
        ((32, 32, 32), "ZYX", (8, 8, 8)),
    ],
)
def test_compute_mean_std_transform_welford(tmp_path, shape, axes, patch_size):
    """Test that mean and std are computed and correctly added to the configuration
    and transform."""
    n_files = 10
    files = []
    array = np.random.randint(0, np.iinfo(np.uint16).max, (n_files, *shape))

    for i in range(n_files):
        file = tmp_path / f"array{i}.tif"
        tifffile.imwrite(file, array[i])
        files.append(file)

    array = array[:, np.newaxis, ...] if "C" not in axes else array

    # create config
    config_dict = {
        "data_type": SupportedData.TIFF.value,
        "patch_size": patch_size,
        "axes": axes,
    }
    config = DataConfig(**config_dict)

    # create dataset
    dataset = PathIterableDataset(
        data_config=config, src_files=files, read_source_func=read_tiff
    )

    # define axes for mean and std computation
    stats_axes = tuple(np.delete(np.arange(array.ndim), 1))

    assert np.allclose(array.mean(axis=stats_axes), dataset.data_config.image_means)
    assert np.allclose(array.std(axis=stats_axes), dataset.data_config.image_stds)


@pytest.mark.parametrize(
    "shape, axes, patch_size",
    [
        ((32, 32), "YX", (8, 8)),
        ((2, 32, 32), "CYX", (8, 8)),
        ((32, 32, 32), "ZYX", (8, 8, 8)),
    ],
)
def test_compute_mean_std_transform_welford_with_targets(
    tmp_path, shape, axes, patch_size
):
    """Test that mean and std are computed and correctly added to the configuration
    and transform."""
    n_files = 10
    files = []
    target_files = []
    array = np.random.randint(0, np.iinfo(np.uint16).max, (n_files, *shape))
    target_array = np.random.randint(0, np.iinfo(np.uint16).max, (n_files, *shape))

    for i in range(n_files):
        file = tmp_path / "images" / f"array{i}.tif"
        target_file = tmp_path / "targets" / f"array{i}.tif"
        os.makedirs(file.parent, exist_ok=True)
        os.makedirs(target_file.parent, exist_ok=True)
        tifffile.imwrite(file, array[i])
        tifffile.imwrite(target_file, target_array[i])
        files.append(file)
        target_files.append(target_file)

    array = array[:, np.newaxis, ...] if "C" not in axes else array
    target_array = target_array[:, np.newaxis, ...] if "C" not in axes else target_array

    # create config
    config_dict = {
        "data_type": SupportedData.TIFF.value,
        "patch_size": patch_size,
        "axes": axes,
    }
    config = DataConfig(**config_dict)

    # create dataset
    dataset = PathIterableDataset(
        data_config=config,
        src_files=files,
        target_files=target_files,
        read_source_func=read_tiff,
    )

    # define axes for mean and std computation
    stats_axes = tuple(np.delete(np.arange(array.ndim), 1))

    assert np.allclose(
        target_array.mean(axis=stats_axes), dataset.data_config.target_means
    )
    assert np.allclose(
        target_array.std(axis=stats_axes), dataset.data_config.target_stds
    )



import pytest

from careamics.config.architectures import LVAEModel
from careamics.config.support import SupportedActivation

pytestmark = pytest.mark.lvae


def test_instantiation():
    """Test that LVAEModel can be instantiated."""
    model_params = {
        "architecture": "LVAE",
    }

    # instantiate model
    LVAEModel(**model_params)


def test_architecture_missing():
    """Test that LVAEModel requires architecture."""
    model_params = {
        "input_shape": 64,
    }

    with pytest.raises(ValueError):
        LVAEModel(**model_params)


@pytest.mark.parametrize("encoder_n_filters", [8, 16, 32, 96, 128])
def test_encoder_n_filters(encoder_n_filters: int):
    """Test that LVAEModel accepts num_channels_init as an even number and
    minimum 8."""
    model_params = {"architecture": "LVAE", "encoder_n_filters": encoder_n_filters}

    # instantiate model
    LVAEModel(**model_params)


@pytest.mark.parametrize("decoder_n_filters", [8, 16, 32, 96, 128])
def test_decoder_n_filters(decoder_n_filters: int):
    """Test that LVAEModel accepts num_channels_init as an even number and
    minimum 8."""
    model_params = {"architecture": "LVAE", "decoder_n_filters": decoder_n_filters}

    # instantiate model
    LVAEModel(**model_params)


@pytest.mark.parametrize("n_filters", [2, 17, 127])
def test_wrong_num_filters(n_filters: int):
    """Test that wrong num_channels_init causes an error."""
    model_params = {"architecture": "LVAE", "encoder_n_filters": n_filters}
    with pytest.raises(ValueError):
        LVAEModel(**model_params)

    model_params = {"architecture": "LVAE", "decoder_n_filters": n_filters}
    with pytest.raises(ValueError):
        LVAEModel(**model_params)


def test_activations():
    """Test that LVAEModel accepts all activations."""
    for act in SupportedActivation:
        model_params = {
            "architecture": "LVAE",
            "nonlinearity": act.value,
        }

        # instantiate model
        LVAEModel(**model_params)


def test_all_activations_are_supported():
    """Test that all activations defined in the Literal are supported."""
    # list of supported activations
    activations = list(SupportedActivation)

    # Algorithm json schema
    schema = LVAEModel.model_json_schema()

    # check that all activations are supported
    for act in schema["properties"]["nonlinearity"]["enum"]:
        assert act in activations


def test_activation_wrong_values():
    """Test that wrong values are not accepted."""
    model_params = {
        "architecture": "LVAE",
        "nonlinearity": "wrong",
    }

    with pytest.raises(ValueError):
        LVAEModel(**model_params)


def test_parameters_wrong_values_by_assigment():
    """Test that wrong values are not accepted through assignment."""
    model_params = {
        "architecture": "LVAE",
        "z_dims": (128, 128, 128),
        "multiscale_count": 2,
        "encoder_n_filters": 32,
    }
    model = LVAEModel(**model_params)

    # z_dims
    model.z_dims = model_params["z_dims"]
    with pytest.raises(ValueError):
        model.depth = -1

    # number of channels in the encoder
    model.encoder_n_filters = model_params["encoder_n_filters"]
    with pytest.raises(ValueError):
        model.encoder_n_filters = 2


def test_model_dump():
    """Test that default values are excluded from model dump."""
    model_params = {
        "architecture": "LVAE",  # default value
        "z_dims": (128, 128, 128, 128),  # default value
        "nonlinearity": "ReLU",  # non-default value
        "decoder_n_filters": 32,  # non-default value
    }
    model = LVAEModel(**model_params)

    # dump model
    model_dict = model.model_dump(exclude_defaults=True)

    # check that default values are excluded except the architecture
    assert "architecture" not in model_dict
    assert len(model_dict) == 3  # TODO not sure it's hardcoded?

    # check that we get all the optional values with the exclude_defaults flag
    model_dict = model.model_dump(exclude_defaults=False)
    assert len(model_dict) == len(dict(model)) - 1
